"""
Module to provide context to markdown transforms.
"""

import logging
from dataclasses import dataclass
from typing import List, Optional

from typing_extensions import Protocol

from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.transform_gfm.transform_state import TransformState

POGGER = ParserLogger(logging.getLogger(__name__))


@dataclass
class IndentAdjustment:
    """
    Class to hold indent adjustments.
    """

    adjustment: int = 0


# pylint: disable=too-few-public-methods
class MarkdownTransformContext:
    """
    Context to preserve state for the markdown transform.
    """

    def __init__(self) -> None:
        self.block_stack: List[MarkdownToken] = []
        self.container_token_stack: List[MarkdownToken] = []
        self.original_container_token_stack: List[MarkdownToken] = []
        self.container_token_indents: List[IndentAdjustment] = []


# pylint: enable=too-few-public-methods


# pylint: disable=too-few-public-methods
class StartMarkdownTokenTransformProtocol(Protocol):
    """
    Protocol to call for the start token of a markdown transform.
    """

    def __call__(  # noqa: E704
        self,
        context: "MarkdownTransformContext",
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str: ...  # pragma: no cover


# pylint: enable=too-few-public-methods


# pylint: disable=too-few-public-methods
class EndMarkdownTokenTransformProtocol(Protocol):
    """
    Protocol to allow registration of handlers for the markdown transform.
    """

    def __call__(  # noqa: E704
        self,
        context: "MarkdownTransformContext",
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
        next_token: Optional[MarkdownToken],
    ) -> str: ...  # pragma: no cover


# pylint: enable=too-few-public-methods


# pylint: disable=too-few-public-methods
class StartMarkdownContainerTokenTransformProtocol(Protocol):
    """
    Protocol to call for the start container token of a markdown transform.
    """

    # pylint: disable=too-many-arguments
    def __call__(  # noqa: E704
        self,
        context: "MarkdownTransformContext",
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
        next_token: Optional[MarkdownToken],
        transformed_data: str,
    ) -> str: ...  # pragma: no cover

    # pylint: enable=too-many-arguments


# pylint: enable=too-few-public-methods


# pylint: disable=too-few-public-methods
class EndMarkdownContainerTokenTransformProtocol(Protocol):
    """
    Protocol to allow registration of handlers for the container tokens of a markdown transform.
    """

    def __call__(  # noqa: E704
        self,
        context: "MarkdownTransformContext",
        current_token: MarkdownToken,
        actual_tokens: List[MarkdownToken],
        token_index: int,
    ) -> str: ...  # pragma: no cover


# pylint: enable=too-few-public-methods


# pylint: disable=too-few-public-methods
class RegisterMarkdownTransformHandlersProtocol(Protocol):
    """
    Protocol to allow registration of handlers for the markdown transform.
    """

    def __call__(  # noqa: E704
        self,
        type_to_register: type,
        start_function: StartMarkdownTokenTransformProtocol,
        end_function: Optional[EndMarkdownTokenTransformProtocol],
    ) -> None: ...  # pragma: no cover


# pylint: enable=too-few-public-methods


# pylint: disable=too-few-public-methods
class StartHtmlTokenTransformProtocol(Protocol):
    """
    Protocol to call for the start token of a html transform.
    """

    def __call__(  # noqa: E704
        self,
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str: ...  # pragma: no cover


# pylint: enable=too-few-public-methods


# pylint: disable=too-few-public-methods
class EndHtmlTokenTransformProtocol(Protocol):
    """
    Protocol to call for the end token or a html transform.
    """

    def __call__(  # noqa: E704
        self,
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str: ...  # pragma: no cover


# pylint: enable=too-few-public-methods


# pylint: disable=too-few-public-methods
class RegisterHtmlTransformHandlersProtocol(Protocol):
    """
    Protocol to allow registration of handlers for the html transform.
    """

    def __call__(  # noqa: E704
        self,
        type_to_register: type,
        start_function: StartHtmlTokenTransformProtocol,
        end_function: Optional[EndHtmlTokenTransformProtocol],
    ) -> None: ...  # pragma: no cover


# pylint: enable=too-few-public-methods



"""
Module to hold the token to fix and how to fix it.
"""

from dataclasses import dataclass
from typing import Union

from pymarkdown.tokens.markdown_token import MarkdownToken


@dataclass(frozen=True)
class FixTokenRecord:
    """
    Class to hold the token to fix and how to fix it.
    """

    token_to_fix: MarkdownToken
    plugin_id: str
    plugin_action: str
    field_name: str
    field_value: Union[str, int]



"""
Module to provide for encapsulation on a group of container indices.
"""

from dataclasses import dataclass


@dataclass(frozen=True)
class ContainerIndices:
    """
    Class to provide for encapsulation on a group of container indices.
    """

    ulist_index: int
    olist_index: int
    block_index: int

    def __str__(self) -> str:
        return (
            f"{{ContainerIndices:ulist_index:{self.ulist_index};olist_index:{self.olist_index};"
            + f"block_index:{self.block_index}}}"
        )



"""
Module to provide transformations for a new list item.
"""

import logging
from typing import Optional, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.new_list_item_markdown_token import NewListItemMarkdownToken
from pymarkdown.transform_markdown.markdown_transform_context import (
    MarkdownTransformContext,
)
from pymarkdown.transform_markdown.transform_list_block import TransformListBlock

POGGER = ParserLogger(logging.getLogger(__name__))

# pylint: disable=too-few-public-methods


class TransformNewListItem:
    """
    Class to provide transformations for a new list item.
    """

    # pylint: disable=too-many-locals
    @staticmethod
    def rehydrate_next_list_item(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
        next_token: Optional[MarkdownToken],
        transformed_data: str,
    ) -> str:
        """
        Rehydrate the next list item token.
        """
        assert next_token is not None, "If there is a list start, must have a list end."
        assert (
            previous_token is not None
        ), "If this is a new list item, must have a list start."
        assert context.container_token_stack[
            -1
        ].is_list_start, "Must occured within a list."
        # assert isinstance(context.container_token_stack[-1], ListStartMarkdownToken)
        list_start_token = cast(
            ListStartMarkdownToken, context.container_token_stack[-1]
        )

        current_list_token = cast(NewListItemMarkdownToken, current_token)

        POGGER.debug("__rehydrate_next_list_item")
        list_start_token.adjust_for_new_list_item(current_list_token)

        (
            adjustment_since_newline,
            extracted_whitespace,
        ) = TransformListBlock.adjust_whitespace_for_block_quote(
            transformed_data, current_list_token.extracted_whitespace
        )
        POGGER.debug(f"rnli->adjustment_since_newline>:{adjustment_since_newline}:")
        POGGER.debug(f"rnli->extracted_whitespace>:{extracted_whitespace}:")

        did_container_start_midline = False
        had_weird_block_quote_in_list = False
        (
            previous_indent,
            extracted_whitespace2,
            _,
            post_adjust_whitespace,
            did_container_start_midline,
            had_weird_block_quote_in_list,
        ) = TransformListBlock.rehydrate_list_start_previous_token(
            context,
            current_list_token,
            previous_token,
            extracted_whitespace,
        )
        # else:
        #     previous_indent, post_adjust_whitespace = (0, None)
        POGGER.debug(f">>previous_indent>>{previous_indent}<<")
        POGGER.debug(f">>extracted_whitespace2>>{extracted_whitespace2}<<")
        POGGER.debug(f">>post_adjust_whitespace>>{post_adjust_whitespace}<<")
        POGGER.debug(
            f">>had_weird_block_quote_in_list>>{had_weird_block_quote_in_list}<<"
        )

        adjustment_since_newline = (
            TransformNewListItem.__recalc_adjustment_since_newline(
                context, adjustment_since_newline, current_token
            )
        )
        # assert len(post_adjust_whitespace) == adjustment_since_newline

        whitespace_to_use = (
            post_adjust_whitespace
            if did_container_start_midline or had_weird_block_quote_in_list
            else extracted_whitespace
        )

        POGGER.debug(f"rnli->whitespace_to_use>:{whitespace_to_use}:")
        POGGER.debug(f"rnli->adjustment_since_newline>:{adjustment_since_newline}:")
        POGGER.debug(f"rnli->extracted_whitespace>:{extracted_whitespace}:")
        start_sequence = (
            f"{whitespace_to_use}{current_list_token.list_start_content}"
            + f"{list_start_token.list_start_sequence}"
        )

        POGGER.debug(f"rnli->start_sequence>:{start_sequence}:")
        start_sequence = (
            TransformNewListItem.__rehydrate_next_list_item_blank_line(
                start_sequence, current_list_token, next_token
            )
            if next_token.is_blank_line
            else (
                TransformNewListItem.__rehydrate_next_list_item_not_blank_line(
                    context,
                    start_sequence,
                    did_container_start_midline,
                    adjustment_since_newline,
                    had_weird_block_quote_in_list,
                    next_token,
                )
            )
        )
        POGGER.debug(f"rnli->start_sequence>:{start_sequence}:")
        return start_sequence

    # pylint: enable=too-many-locals

    # pylint: disable=too-many-arguments
    @staticmethod
    def __rehydrate_next_list_item_not_blank_line(
        context: MarkdownTransformContext,
        start_sequence: str,
        did_container_start_midline: bool,
        adjustment_since_newline: int,
        had_weird_block_quote_in_list: bool,
        next_token: MarkdownToken,
    ) -> str:
        POGGER.debug("__rehydrate_next_list_item_not_blank_line")
        # POGGER.debug(f"start_sequence={start_sequence}=")
        # POGGER.debug(f"did_container_start_midline={did_container_start_midline}=")
        # POGGER.debug(f"adjustment_since_newline={adjustment_since_newline}=")

        assert context.container_token_stack[
            -1
        ].is_list_start, "If this is a new list item, must have a list start."
        list_start_token = cast(
            ListStartMarkdownToken, context.container_token_stack[-1]
        )
        # assert isinstance(context.container_token_stack[-1], ListStartMarkdownToken)

        if did_container_start_midline:
            POGGER.debug("did start midline")
            # POGGER.debug(f"next_token:{ParserHelper.make_value_visible(next_token)}")
            project_indent_level = list_start_token.indent_level
            if next_token and next_token.is_block_quote_start:
                next_block_token = cast(BlockQuoteMarkdownToken, next_token)
                next_block_quote_leading_space = (
                    next_block_token.calculate_next_bleading_space_part(
                        increment_index=False
                    )
                )
                # POGGER.debug(
                #     f"did start midline:next_block_quote_leading_space:{next_block_quote_leading_space}:"
                # )
                ex_whitespace, _ = ParserHelper.extract_spaces_verified(
                    next_block_quote_leading_space, 0
                )
                # POGGER.debug(f"did start midline:ab:{ex_whitespace}:")
                project_indent_level -= ex_whitespace
            start_sequence = start_sequence.ljust(project_indent_level, " ")
        else:
            POGGER.debug("did not start midline")
            calculated_indent = list_start_token.indent_level - adjustment_since_newline
            POGGER.debug(
                f"calculated_indent:{calculated_indent} = indent_level:{list_start_token.indent_level} - adjustment_since_newline:{adjustment_since_newline}"
            )
            POGGER.debug(
                f"had_weird_block_quote_in_list:{had_weird_block_quote_in_list}"
            )
            if had_weird_block_quote_in_list:
                POGGER.debug(f"calculated_indent:{calculated_indent}")
                calculated_indent += 2
                POGGER.debug(f"calculated_indent:{calculated_indent}")
            POGGER.debug(
                f"rnli->calculated_indent={calculated_indent} = "
                + f"indent_level={list_start_token.indent_level} - "
                + f"adjustment_since_newline={adjustment_since_newline}"
            )
            POGGER.debug(f"start_sequence:{start_sequence}")
            start_sequence = start_sequence.ljust(calculated_indent, " ")

            # TODO This is a kludge.  The calc_indent is not properly computed.
            if not start_sequence.endswith(" "):
                start_sequence = f"{start_sequence} "
            POGGER.debug(f"start_sequence:{start_sequence}")
        return start_sequence

    # pylint: enable=too-many-arguments

    @staticmethod
    def __rehydrate_next_list_item_blank_line(
        start_sequence: str,
        current_token: NewListItemMarkdownToken,
        next_token: MarkdownToken,
    ) -> str:
        POGGER.debug(f">>next_token.column_number>>:{next_token.column_number}:<<")
        POGGER.debug(
            f">>current_token.column_number>>:{current_token.column_number}:<<"
        )
        start_content_length = 1
        if current_token.list_start_content:
            start_content_length += len(current_token.list_start_content)
        new_column_number = (
            next_token.column_number
            - current_token.column_number
            - start_content_length
        )
        start_sequence += ParserHelper.repeat_string(" ", new_column_number)
        return start_sequence

    @staticmethod
    def __recalc_adjustment_since_newline(
        context: MarkdownTransformContext,
        adjustment_since_newline: int,
        current_token: MarkdownToken,
    ) -> int:
        POGGER.debug(
            f"rnli->container_token_stack>:{ParserHelper.make_value_visible(context.container_token_stack)}:"
        )
        stack_index = len(context.container_token_stack) - 1
        found_block_quote_token: Optional[BlockQuoteMarkdownToken] = None
        while stack_index >= 0:
            if context.container_token_stack[stack_index].is_block_quote_start:
                found_block_quote_token = cast(
                    BlockQuoteMarkdownToken, context.container_token_stack[stack_index]
                )
                break
            stack_index -= 1
        POGGER.debug(
            f"rnli->found_block_quote_token>:{ParserHelper.make_value_visible(found_block_quote_token)}:"
        )
        if found_block_quote_token:
            line_number_delta = (
                current_token.line_number - found_block_quote_token.line_number
            )
            line_number_delta -= context.container_token_indents[stack_index].adjustment
            assert found_block_quote_token.bleading_spaces is not None
            split_leading_spaces = found_block_quote_token.bleading_spaces.split(
                ParserHelper.newline_character
            )
            leading_space = split_leading_spaces[line_number_delta]

            POGGER.debug(f"rnli->leading_space>:{leading_space}:")
            adjustment_since_newline = len(leading_space)
        return adjustment_since_newline


# pylint: enable=too-few-public-methods



"""
Module to implement a plugin that ensures the code blocks maintain a consistent style.
"""

from typing import List, Optional, Tuple, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.plugin_manager.plugin_details import (
    PluginDetails,
    PluginDetailsV3,
    QueryConfigItem,
)
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.blank_line_markdown_token import BlankLineMarkdownToken
from pymarkdown.tokens.fenced_code_block_markdown_token import (
    FencedCodeBlockMarkdownToken,
)
from pymarkdown.tokens.indented_code_block_markdown_token import (
    IndentedCodeBlockMarkdownToken,
)
from pymarkdown.tokens.markdown_token import EndMarkdownToken, MarkdownToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken


class RuleMd046(RulePlugin):
    """
    Class to implement a plugin that ensures the code blocks maintain a consistent style.
    """

    __consistent_style = "consistent"
    __fenced_style = "fenced"
    __indented_style = "indented"
    __valid_styles = [
        __consistent_style,
        __fenced_style,
        __indented_style,
    ]

    def __init__(self) -> None:
        super().__init__()
        self.__style_type: str = ""
        self.__actual_style_type: str = ""
        self.__start_fix_token: Optional[MarkdownToken] = None
        self.__inner_fix_token: Optional[MarkdownToken] = None
        self.__token_before_start_fix_token: Optional[MarkdownToken] = None
        self.__last_token: Optional[MarkdownToken] = None

    def get_details(self) -> PluginDetails:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV3(
            plugin_name="code-block-style",
            plugin_id="MD046",
            plugin_enabled_by_default=True,
            plugin_description="Code block style",
            plugin_version="0.7.0",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md046.md",
            plugin_configuration="style",
            plugin_supports_fix=True,
        )

    @classmethod
    def __validate_configuration_style(cls, found_value: str) -> None:
        if found_value not in RuleMd046.__valid_styles:
            raise ValueError(f"Allowable values: {RuleMd046.__valid_styles}")

    def initialize_from_config(self) -> None:
        """
        Event to allow the plugin to load configuration information.
        """
        self.__style_type = self.plugin_configuration.get_string_property_with_default(
            "style",
            RuleMd046.__consistent_style,
            valid_value_fn=self.__validate_configuration_style,
        )

    def query_config(self) -> List[QueryConfigItem]:
        """
        Query to find out the configuration that the rule is using.
        """
        return [
            QueryConfigItem("style", self.__style_type),
        ]

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__actual_style_type = (
            self.__style_type
            if self.__style_type != RuleMd046.__consistent_style
            else ""
        )
        self.__last_token = None
        self.__start_fix_token = None
        self.__inner_fix_token = None

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        if self.__start_fix_token:
            if token.is_end_token:
                end_token = cast(EndMarkdownToken, token)
                assert end_token.start_markdown_token == self.__start_fix_token
                self.__fix(context, end_token)
                self.__token_before_start_fix_token = None
                self.__start_fix_token = None
                self.__inner_fix_token = None
            else:
                assert (
                    token.is_text
                ), "Once into a code block, should either be a text token or an end token."
                assert self.__inner_fix_token is None
                self.__inner_fix_token = token
        elif token.is_code_block:
            current_style = (
                RuleMd046.__fenced_style
                if token.is_fenced_code_block
                else RuleMd046.__indented_style
            )
            if not self.__actual_style_type:
                self.__actual_style_type = current_style
            if self.__actual_style_type != current_style:
                if context.in_fix_mode:
                    self.__start_fix_token = token
                    self.__token_before_start_fix_token = self.__last_token
                else:
                    extra_data = (
                        f"Expected: {self.__actual_style_type}; Actual: {current_style}"
                    )
                    self.report_next_token_error(
                        context, token, extra_error_information=extra_data
                    )
        self.__last_token = token

    def __create_new_fenced_tokens(
        self,
    ) -> Tuple[List[MarkdownToken], EndMarkdownToken]:
        collected_count = 3
        new_fenced_start_token = FencedCodeBlockMarkdownToken(
            fence_character="`",
            fence_count=collected_count,
            extracted_text="",
            pre_extracted_text="",
            text_after_extracted_text="",
            pre_text_after_extracted_text="",
            extracted_whitespace="",
            extracted_whitespace_before_info_string="",
            position_marker=PositionMarker(0, 0, ""),
        )
        extracted_spaces = ""
        extra_end_data = f"{extracted_spaces}:{collected_count}"
        new_end_token = EndMarkdownToken(
            new_fenced_start_token.token_name,
            extracted_whitespace="",
            extra_end_data=extra_end_data,
            start_markdown_token=new_fenced_start_token,
            was_forced=False,
        )
        replacement_tokens: List[MarkdownToken] = [new_fenced_start_token]
        return replacement_tokens, new_end_token

    def __create_new_indented_tokens(
        self,
    ) -> Tuple[List[MarkdownToken], EndMarkdownToken]:
        new_indented_start_token = IndentedCodeBlockMarkdownToken(
            extracted_whitespace="    ", line_number=0, column_number=0
        )
        new_end_token = EndMarkdownToken(
            new_indented_start_token.token_name,
            extracted_whitespace="",
            extra_end_data=None,
            start_markdown_token=new_indented_start_token,
            was_forced=False,
        )

        if self.__inner_fix_token:
            inner_text_token = cast(TextMarkdownToken, self.__inner_fix_token)
            newlines_in_inner_text_token = ParserHelper.count_newlines_in_text(
                inner_text_token.token_text
            )
            for _ in range(newlines_in_inner_text_token):
                new_indented_start_token.add_indented_whitespace("    ")
        replacement_tokens = [cast(MarkdownToken, new_indented_start_token)]
        return replacement_tokens, new_end_token

    def __fix(self, context: PluginScanContext, end_fix_token: MarkdownToken) -> None:
        if self.__actual_style_type == RuleMd046.__fenced_style:
            replacement_tokens, new_end_token = self.__create_new_fenced_tokens()
        else:
            replacement_tokens, new_end_token = self.__create_new_indented_tokens()

        if self.__inner_fix_token:
            replacement_tokens.append(self.__inner_fix_token)
        replacement_tokens.append(new_end_token)

        if (
            self.__actual_style_type == RuleMd046.__indented_style
            and self.__token_before_start_fix_token
            and self.__token_before_start_fix_token.is_paragraph_end
        ):
            replacement_tokens.insert(
                0,
                BlankLineMarkdownToken(
                    extracted_whitespace="", position_marker=PositionMarker(0, 0, "")
                ),
            )

        assert self.__start_fix_token is not None
        self.register_replace_tokens_request(
            context, self.__start_fix_token, end_fix_token, replacement_tokens
        )
        self.__start_fix_token = None
        self.__inner_fix_token = None



"""
Module to hold information regarding a line fix that was made.
"""

from dataclasses import dataclass


@dataclass(frozen=True)
class FixLineRecord:
    """
    Class to hold information regarding a line fix that was made.
    """

    source: str
    line_number: int
    plugin_id: str



"""
Module to provide for a simplified way to scan for files.
"""

import argparse
import glob
import logging
import os
from typing import List, Optional, Set, Tuple

from typing_extensions import Protocol

LOGGER = logging.getLogger(__name__)


# pylint: disable=too-few-public-methods
class ApplicationFileScannerOutputProtocol(Protocol):
    """
    Protocol to provide for redirection of output (standard or error).
    """

    def __call__(  # noqa: E704
        self,
        output_string: str,
    ) -> None: ...  # pragma: no cover


# pylint: enable=too-few-public-methods


class ApplicationFileScanner:
    """
    Class to provide for a simplified way to scan for files.
    """

    @staticmethod
    def determine_files_to_scan_with_args(
        args: argparse.Namespace,
        handle_output: ApplicationFileScannerOutputProtocol,
        handle_error: ApplicationFileScannerOutputProtocol,
        paths_to_exclude: Optional[List[str]] = None,
    ) -> Tuple[List[str], bool, bool]:
        """
        Determine the files to scan based on the arguments provided by the `add_default_command_line_arguments` function.
        """
        return ApplicationFileScanner.determine_files_to_scan(
            args.paths,
            paths_to_exclude if paths_to_exclude is not None else args.path_exclusions,
            args.recurse_directories,
            args.alternate_extensions,
            args.list_files,
            handle_output,
            handle_error,
        )

    # pylint: disable=too-many-arguments
    @staticmethod
    def determine_files_to_scan(
        eligible_paths: List[str],
        paths_to_exclude: List[str],
        recurse_directories: bool,
        eligible_extensions: str,
        only_list_files: bool,
        handle_output: ApplicationFileScannerOutputProtocol,
        handle_error: ApplicationFileScannerOutputProtocol,
    ) -> Tuple[List[str], bool, bool]:
        """
        Determine the files to scan, and how to scan for those files.
        """
        split_eligible_extensions: List[str] = eligible_extensions.split(",")

        did_error_scanning_files = False
        files_to_parse: Set[str] = set()
        for next_path in eligible_paths:
            did_error_scanning_files = (
                ApplicationFileScanner.__process_next_eligible_path(
                    next_path,
                    files_to_parse,
                    recurse_directories,
                    split_eligible_extensions,
                    handle_error,
                )
            )
            if did_error_scanning_files:
                break

        if paths_to_exclude:
            for next_path in paths_to_exclude:
                ApplicationFileScanner.__process_next_exclude_path(
                    files_to_parse, next_path
                )

        sorted_files_to_parse = sorted(files_to_parse)
        LOGGER.info("Number of files found: %d", len(sorted_files_to_parse))
        did_only_list_files = ApplicationFileScanner.__handle_main_list_files(
            only_list_files, sorted_files_to_parse, handle_output, handle_error
        )
        return sorted_files_to_parse, did_error_scanning_files, did_only_list_files

    # pylint: enable=too-many-arguments

    @staticmethod
    def __process_next_exclude_path(eligible_paths: Set[str], next_path: str) -> None:
        if "*" in next_path or "?" in next_path:
            globbed_paths = glob.glob(next_path, recursive=True)
            for next_globbed_path in globbed_paths:
                LOGGER.debug(
                    "Provided globbed path '%s' includes the file '%s'. Removing from list if present.",
                    next_path,
                    next_globbed_path,
                )
                if next_globbed_path in eligible_paths:
                    eligible_paths.remove(next_globbed_path)
        elif os.path.exists(next_path):
            if os.path.isdir(next_path):
                if not next_path.endswith(os.sep):
                    next_path += os.sep
                paths_to_remove = {x for x in eligible_paths if x.startswith(next_path)}
                for path in paths_to_remove:
                    LOGGER.debug(
                        "Provided path '%s' is a directory. Removing '%s' from list if present.",
                        next_path,
                        path,
                    )
                    eligible_paths.remove(path)
            elif os.path.isfile(next_path) and next_path in eligible_paths:
                LOGGER.debug(
                    "Provided exclude path '%s' is a file. Removing from list if present.",
                    next_path,
                )
                eligible_paths.remove(next_path)

    @staticmethod
    def __process_next_eligible_path(
        next_path: str,
        files_to_parse: Set[str],
        recurse_directories: bool,
        split_eligible_extensions: List[str],
        handle_error: ApplicationFileScannerOutputProtocol,
    ) -> bool:
        did_error_scanning_files = False
        if "*" in next_path or "?" in next_path:
            globbed_paths = glob.glob(next_path, recursive=True)
            if not globbed_paths:
                handle_error(
                    f"Provided glob path '{next_path}' did not match any files."
                )
                did_error_scanning_files = True
            for next_globbed_path in globbed_paths:
                ApplicationFileScanner.__process_next_path(
                    next_globbed_path,
                    files_to_parse,
                    recurse_directories,
                    split_eligible_extensions,
                    handle_error,
                )
        elif not ApplicationFileScanner.__process_next_path(
            next_path,
            files_to_parse,
            recurse_directories,
            split_eligible_extensions,
            handle_error,
        ):
            did_error_scanning_files = True
        return did_error_scanning_files

    @staticmethod
    def __process_next_path(
        next_path: str,
        files_to_parse: Set[str],
        recurse_directories: bool,
        eligible_extensions: List[str],
        handle_error: ApplicationFileScannerOutputProtocol,
    ) -> bool:
        did_find_any = False
        LOGGER.info("Determining files to scan for path '%s'.", next_path)
        if not os.path.exists(next_path):
            handle_error(f"Provided path '{next_path}' does not exist.")
            LOGGER.debug("Provided path '%s' does not exist.", next_path)
        elif os.path.isdir(next_path):
            ApplicationFileScanner.__process_next_path_directory(
                next_path, files_to_parse, recurse_directories, eligible_extensions
            )
            did_find_any = True
        elif ApplicationFileScanner.__is_file_eligible_to_scan(
            next_path, eligible_extensions
        ):
            LOGGER.debug(
                "Provided path '%s' is a valid file. Adding.",
                next_path,
            )
            normalized_path = (
                next_path.replace(os.altsep, os.sep) if os.altsep else next_path
            )
            files_to_parse.add(normalized_path)
            did_find_any = True
        else:
            LOGGER.debug(
                "Provided path '%s' is not a valid file. Skipping.",
                next_path,
            )
            handle_error(
                f"Provided file path '{next_path}' is not a valid file. Skipping."
            )
        return did_find_any

    @staticmethod
    def __process_next_path_directory(
        next_path: str,
        files_to_parse: Set[str],
        recurse_directories: bool,
        eligible_extensions: List[str],
    ) -> None:
        LOGGER.debug("Provided path '%s' is a directory. Walking directory.", next_path)
        normalized_next_path = (
            next_path.replace(os.altsep, os.sep) if os.altsep else next_path
        )
        for root, _, files in os.walk(normalized_next_path):
            normalized_root = root.replace(os.altsep, os.sep) if os.altsep else root
            if not recurse_directories and normalized_root != normalized_next_path:
                continue
            normalized_root = (
                normalized_root[:-1]
                if normalized_root.endswith(os.sep)
                else normalized_root
            )
            for file in files:
                rooted_file_path = f"{normalized_root}{os.sep}{file}"
                if ApplicationFileScanner.__is_file_eligible_to_scan(
                    rooted_file_path, eligible_extensions
                ):
                    files_to_parse.add(rooted_file_path)

    @staticmethod
    def __is_file_eligible_to_scan(
        path_to_test: str, eligible_extensions: List[str]
    ) -> bool:
        """
        Determine if the presented path is one that we want to scan.
        """
        return os.path.isfile(path_to_test) and any(
            path_to_test.endswith(next_extension)
            for next_extension in eligible_extensions
        )

    # pylint: disable=too-many-arguments
    @staticmethod
    def add_default_command_line_arguments(
        parser_to_add_to: argparse.ArgumentParser,
        extension_to_look_for: str,
        file_type_name: Optional[str] = None,
        show_list_files: bool = True,
        show_recurse_directories: bool = True,
        show_alternate_extensions: bool = True,
        show_exclusions: bool = True,
    ) -> None:
        """
        Add a set of default command line arguments to an argparse styled command line.
        """
        if argument_error := ApplicationFileScanner.__is_valid_extension(
            extension_to_look_for
        ):
            raise argparse.ArgumentTypeError(
                f"Extension '{extension_to_look_for}' is not a valid extension: {argument_error}"
            )

        specific_file_type_name = ""
        if file_type_name is not None:
            if file_type_name := file_type_name.strip():
                specific_file_type_name = f"{file_type_name} "

        if show_list_files:
            parser_to_add_to.add_argument(
                "-l",
                "--list-files",
                dest="list_files",
                action="store_true",
                default=False,
                help=f"list any eligible {specific_file_type_name}files found on the specified paths and exit",
            )

        if show_recurse_directories:
            parser_to_add_to.add_argument(
                "-r",
                "--recurse",
                dest="recurse_directories",
                action="store_true",
                default=False,
                help="recursively traverse any found directories for matching files",
            )

        if show_alternate_extensions:
            parser_to_add_to.add_argument(
                "-ae",
                "--alternate-extensions",
                dest="alternate_extensions",
                action="store",
                default=extension_to_look_for,
                type=ApplicationFileScanner.is_valid_comma_separated_extension_list,
                help="provide an alternate set of file extensions to match against",
            )

        if show_exclusions:
            parser_to_add_to.add_argument(
                "-e",
                "--exclude",
                dest="path_exclusions",
                action="append",
                type=str,
                help="one or more paths to exclude from the search. Can be a glob pattern.",
            )

        parser_to_add_to.add_argument(
            "paths",
            metavar="path",
            type=str,
            nargs="+",
            default=None,
            help=f"one or more paths to examine for eligible {specific_file_type_name}files",
        )

    # pylint: enable=too-many-arguments

    @staticmethod
    def __is_valid_extension(possible_extension: str) -> Optional[str]:
        """
        Determine if the parameter is a string that has the form of a valid extension.
        """
        if not possible_extension.startswith("."):
            return f"Extension '{possible_extension}' must start with a period."
        return (
            next(
                (
                    f"Extension '{possible_extension}' must only contain alphanumeric characters after the period."
                    for clean_split_char in clean_split
                    if not clean_split_char.isalnum()
                ),
                None,
            )
            if (clean_split := possible_extension[1:])
            else f"Extension '{possible_extension}' must have at least one character after the period."
        )

    @staticmethod
    def is_valid_comma_separated_extension_list(argument: str) -> str:
        """
        Function to help argparse verify whether a comma separated list contains valid extensions.
        """
        split_argument = argument.split(",")
        for next_split in split_argument:
            if error_string := ApplicationFileScanner.__is_valid_extension(next_split):
                raise argparse.ArgumentTypeError(error_string)
        return argument.lower()

    @staticmethod
    def __handle_main_list_files(
        only_list_files: bool,
        files_to_scan: List[str],
        handle_output: ApplicationFileScannerOutputProtocol,
        handle_error: ApplicationFileScannerOutputProtocol,
    ) -> bool:
        if only_list_files:
            LOGGER.info("Sending list of files that would have been scanned to stdout.")
            if files_to_scan:
                handle_output("\n".join(files_to_scan))
            else:
                handle_error("No matching files found.")
        return only_list_files



"""
Module to provide for a simple logger wrapper that provides extra
functionality for logging parsing information.
"""

import logging
from typing import Any, List, cast

from pymarkdown.general.parser_helper import ParserHelper


class ParserLoggerException(Exception):
    """
    Custom exception for the ParserLogger class.
    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


class ParserLogger:
    """
    Class to provide for a simple logger wrapper that provides extra
    functionality for logging parsing information.

    To keep things performant, the calls to the underlying logging libraries
    are only done when needed.
    """

    start_range_sequence = "\u8268"
    end_range_sequence = "\u8269"
    blah_sequence = "\u00fe"

    __global_count = 0

    def __init__(self, my_logger: logging.Logger) -> None:
        self.__my_logger = my_logger
        root_logger = logging.getLogger()
        self.__is_info_enabled = root_logger.isEnabledFor(logging.INFO)
        self.__is_debug_enabled = root_logger.isEnabledFor(logging.DEBUG)
        self.__local_count = ParserLogger.__global_count

    def __reset_cache(self) -> None:
        self.__local_count = ParserLogger.__global_count
        root_logger = logging.getLogger()
        self.__is_info_enabled = root_logger.isEnabledFor(logging.INFO)
        self.__is_debug_enabled = root_logger.isEnabledFor(logging.DEBUG)

    @staticmethod
    def sync_on_next_call() -> None:
        """
        Sync the local instance of the logger on the next call.
        """
        ParserLogger.__global_count += 1

    def info(self, log_format: str, *args: Any) -> None:
        """
        Log information at an "INFO" level to the logger.
        """
        if ParserLogger.__global_count != self.__local_count:
            self.__reset_cache()
        if self.__is_info_enabled:
            msg = self.__munge(False, log_format, cast(List[Any], args))
            self.__my_logger.info(msg, stacklevel=2)

    @property
    def is_debug_enabled(self) -> bool:
        """
        Determine whether debug logging is currently enabled.
        """
        return self.__is_debug_enabled

    def debug(self, log_format: str, *args: Any) -> None:
        """
        Log information at a "DEBUG" level to the logger.
        """
        if ParserLogger.__global_count != self.__local_count:
            self.__reset_cache()
        if self.__is_debug_enabled:
            msg = self.__munge(False, log_format, cast(List[Any], args))
            self.__my_logger.debug(msg, stacklevel=2)

    def debug_with_visible_whitespace(self, log_format: str, *args: Any) -> None:
        """
        Log information at a "DEBUG" level to the logger, but replace the
        automatic filtering of any string with make_value_visible to
        using make_whitespace_visible.

        Note: This is seldom used, and does not have a reset check as
        the reset will be hit by one of the other two functions long
        before it gets here.
        """
        if self.__is_debug_enabled:
            msg = self.__munge(True, log_format, cast(List[Any], args))
            self.__my_logger.debug(msg, stacklevel=2)

    def is_enabled_for(self, debug_level: int) -> bool:
        """
        Wrapper method for the logger.isEnabledFor method.
        """
        return self.__my_logger.isEnabledFor(debug_level)

    @classmethod
    def __munge(cls, show_whitespace: bool, log_format: str, args: List[Any]) -> str:
        split_log_format = log_format.split("$")
        split_log_format_length = len(split_log_format)
        args_length = len(args)
        if split_log_format_length != args_length + 1:
            raise ParserLoggerException(
                "The number of $ substitution characters does not equal the number of arguments in the list."
            )

        recipient_array: List[str] = [""] * (split_log_format_length + args_length)
        for next_array_index, _ in enumerate(recipient_array):
            if next_array_index % 2 == 0:
                recipient_array[next_array_index] = split_log_format[
                    int(next_array_index / 2)
                ]
            elif show_whitespace:
                recipient_array[next_array_index] = (
                    ParserHelper.make_whitespace_visible(
                        args[int(next_array_index / 2)]
                    )
                )
            else:
                recipient_array[next_array_index] = ParserHelper.make_value_visible(
                    args[int(next_array_index / 2)]
                )
        formatted_string = "".join(recipient_array)
        if ParserLogger.start_range_sequence in formatted_string:
            formatted_string = formatted_string.replace(
                ParserLogger.start_range_sequence, ""
            )
        if ParserLogger.end_range_sequence in formatted_string:
            formatted_string = formatted_string.replace(
                ParserLogger.end_range_sequence, ""
            )
        return formatted_string



"""
Module to provide constants to deal with extensions.
"""

# pylint: disable=too-few-public-methods


class ExtensionManagerConstants:
    """
    Class to provide constants to deal with extensions.
    """

    EXTENSION_VERSION_NOT_IMPLEMENTED = "0.0.0"

    EXTENSION_INTERFACE_VERSION_BASIC = 1


# pylint: enable=too-few-public-methods



"""
Base class for images and links.
"""

# pylint: disable=too-many-instance-attributes
from typing import Optional, Tuple, Union

from typing_extensions import override

from pymarkdown.links.link_helper_properties import LinkHelperProperties
from pymarkdown.tokens.inline_markdown_token import InlineMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken


class ReferenceMarkdownToken(InlineMarkdownToken):
    """
    Base class for images and links.
    """

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        lhp: LinkHelperProperties,
        token_name: str,
        extra_data: Optional[str],
        text_from_blocks: str,
        line_number: int = 0,
        column_number: int = 0,
        requires_end_token: bool = False,
        can_force_close: bool = True,
    ):
        self.simple_extra_data = extra_data
        if lhp:
            (
                self.__label_type,
                self.__link_uri,
                self.__link_title,
                self.__pre_link_uri,
                self.__pre_link_title,
                self.__ex_label,
                self.__text_from_blocks,
                self.__did_use_angle_start,
                self.__inline_title_bounding_character,
                self.__before_link_whitespace,
                self.__before_title_whitespace,
                self.__after_title_whitespace,
            ) = (
                lhp.label_type,
                lhp.inline_link,
                lhp.inline_title,
                lhp.pre_inline_link,
                lhp.pre_inline_title,
                lhp.ex_label,
                text_from_blocks,
                lhp.did_use_angle_start,
                lhp.bounding_character,
                lhp.before_link_whitespace,
                lhp.before_title_whitespace,
                lhp.after_title_whitespace,
            )
        else:
            (
                self.__label_type,
                self.__link_uri,
                self.__link_title,
                self.__pre_link_uri,
                self.__pre_link_title,
                self.__ex_label,
                self.__text_from_blocks,
                self.__did_use_angle_start,
                self.__inline_title_bounding_character,
                self.__before_link_whitespace,
                self.__before_title_whitespace,
                self.__after_title_whitespace,
            ) = (
                "",
                "",
                "",
                "",
                "",
                "",
                "",
                False,
                "",
                "",
                "",
                "",
            )

        assert self.__link_uri is not None, "This field should be defined."
        assert self.__label_type is not None, "This field should be defined."
        assert self.__pre_link_uri is not None, "This field should be defined."

        part_1, part_2 = self.__compose_extra_data_field(token_name)

        InlineMarkdownToken.__init__(
            self,
            token_name,
            f"{part_1}{part_2}",
            line_number=line_number,
            column_number=column_number,
            requires_end_token=requires_end_token,
            can_force_close=can_force_close,
        )

    # pylint: enable=too-many-arguments

    def __build_extra_data(
        self, extra_data: Optional[str], label_type: str
    ) -> Tuple[str, str]:
        assert self.__link_title is not None, "This field should be defined."
        assert self.__link_uri is not None, "This field should be defined."
        assert extra_data is not None, "This field should be defined."
        part_1 = MarkdownToken.extra_data_separator.join(
            [label_type, self.__link_uri, self.__link_title, extra_data]
        )
        assert (
            self.__inline_title_bounding_character is not None
        ), "This field should be defined."
        assert (
            self.__before_link_whitespace is not None
        ), "This field should be defined."
        assert (
            self.__before_title_whitespace is not None
        ), "This field should be defined."
        assert (
            self.__after_title_whitespace is not None
        ), "This field should be defined."
        assert self.__pre_link_title is not None, "This field should be defined."
        assert self.__ex_label is not None, "This field should be defined."
        assert self.__pre_link_uri is not None, "This field should be defined."
        part_2 = MarkdownToken.extra_data_separator.join(
            [
                self.__pre_link_uri,
                self.__pre_link_title,
                self.__ex_label,
                self.__text_from_blocks,
                str(self.__did_use_angle_start),
                self.__inline_title_bounding_character,
                self.__before_link_whitespace,
                self.__before_title_whitespace,
                self.__after_title_whitespace,
            ]
        )
        return part_1, part_2

    @property
    def label_type(self) -> str:
        """
        Returns the type of label that was used.
        """
        assert self.__label_type is not None, "Label type must be defined."
        return self.__label_type

    @property
    def link_uri(self) -> str:
        """
        Returns the URI for the link itself.
        """
        assert self.__link_uri is not None, "Link uri must be defined."
        return self.__link_uri

    @property
    def active_link_uri(self) -> str:
        """
        Returns the active URI for the link, preferring the __pre_link_uri over the __link_uri.
        """
        active_link = self.__pre_link_uri or self.__link_uri
        assert active_link is not None, "Active link must be defined."
        return active_link

    @property
    def link_title(self) -> Optional[str]:
        """
        Returns the text associated with the link's title.
        """
        return self.__link_title

    @property
    def pre_link_title(self) -> Optional[str]:
        """
        Returns the text associated with the link's title.
        """
        return self.__pre_link_title

    @property
    def active_link_title(self) -> Optional[str]:
        """
        Returns the active text associated with the link's title, preferring the __pre_link_title over the __link_title.
        """
        return self.__pre_link_title or self.__link_title

    @property
    def ex_label(self) -> Optional[str]:
        """
        Returns the text extracted from the blocks of the link, after processing.
        """
        return self.__ex_label

    @property
    def text_from_blocks(self) -> str:
        """
        Returns the text extracted from the blocks of the link, before processing.
        """
        return self.__text_from_blocks

    @property
    def did_use_angle_start(self) -> Optional[bool]:
        """
        Returns a value indicating whether an angle start was used around the URI.
        """
        return self.__did_use_angle_start

    @property
    def inline_title_bounding_character(self) -> Optional[str]:
        """
        Returns the bounding character used for the title.
        """
        return self.__inline_title_bounding_character

    @property
    def before_link_whitespace(self) -> Optional[str]:
        """
        Returns the whitespace extracted before the link.
        """
        return self.__before_link_whitespace

    @property
    def before_title_whitespace(self) -> Optional[str]:
        """
        Returns the whitespace extracted before the title.
        """
        return self.__before_title_whitespace

    @property
    def after_title_whitespace(self) -> Optional[str]:
        """
        Returns the whitespace extracted after the title.
        """
        return self.__after_title_whitespace

    def __compose_extra_data_field(self, token_name: str) -> Tuple[str, str]:
        """
        Compose the object's self.extra_data field from the local object's variables.
        """
        extra_data = (
            f"{self.simple_extra_data}{MarkdownToken.extra_data_separator}"
            if token_name == MarkdownToken._token_inline_image
            else self.simple_extra_data
        )

        # Purposefully split this way to accommodate the extra data
        assert self.__label_type is not None, "Label type must be defined."
        part_1, part_2 = self.__build_extra_data(extra_data, self.__label_type)
        self._set_extra_data(f"{part_1}{part_2}")
        return part_1, part_2

    @override
    def _modify_token(self, field_name: str, field_value: Union[str, int]) -> bool:
        if field_name == "text_from_blocks" and isinstance(field_value, str):
            self.__text_from_blocks = field_value
            self.__compose_extra_data_field(self.token_name)

            return True
        if field_name == "link_title" and isinstance(field_value, str):
            self.__link_title = field_value
            self.__compose_extra_data_field(self.token_name)

            return True
        if field_name == "pre_link_title" and isinstance(field_value, str):
            self.__pre_link_title = field_value
            self.__compose_extra_data_field(self.token_name)

            return True
        return False


# pylint: enable=too-many-instance-attributes



"""
Module to provide an encapsulation of the location within the Markdown document.
"""

from dataclasses import dataclass


@dataclass(frozen=True)
class PositionMarker:
    """
    Class to provide an encapsulation of the location within the Markdown document.
    """

    line_number: int
    index_number: int
    text_to_parse: str
    index_indent: int = 0



"""
Module to provide for the recognition of Markdown tables.
"""

from pymarkdown.extension_manager.extension_impl import ExtensionDetails
from pymarkdown.extension_manager.extension_manager_constants import (
    ExtensionManagerConstants,
)
from pymarkdown.extension_manager.parser_extension import ParserExtension
from pymarkdown.my_application_properties_facade import MyApplicationPropertiesFacade


class MarkdownTablesExtension(ParserExtension):
    """
    Extension to implement the tables extension.
    """

    def get_identifier(self) -> str:
        """
        Get the identifier associated with this extension.
        """
        return "markdown-tables"

    def get_details(self) -> ExtensionDetails:
        """
        Get the details for the extension.
        """
        return ExtensionDetails(
            extension_id=self.get_identifier(),
            extension_name="Markdown Tables",
            extension_description="Allows parsing of Markdown tables.",
            extension_enabled_by_default=False,
            extension_version="0.1.0",
            extension_interface_version=ExtensionManagerConstants.EXTENSION_INTERFACE_VERSION_BASIC,
            extension_url="https://github.github.com/gfm/#tables-extension-",
            extension_configuration=None,
        )

    def apply_configuration(
        self, extension_specific_facade: MyApplicationPropertiesFacade
    ) -> None:
        """
        Apply any configuration required by the extension.
        """
        _ = extension_specific_facade  # pragma: no cover



"""
Module to provide for an encapsulation of the blank line element.
"""

from typing import Optional, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.tokens.leaf_markdown_token import LeafMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.transform_gfm.transform_state import TransformState
from pymarkdown.transform_markdown.markdown_transform_context import (
    MarkdownTransformContext,
    RegisterHtmlTransformHandlersProtocol,
    RegisterMarkdownTransformHandlersProtocol,
)


class BlankLineMarkdownToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the blank line element.
    """

    def __init__(
        self,
        extracted_whitespace: str,
        position_marker: Optional[PositionMarker],
        column_delta: int = 0,
    ) -> None:
        line_number = position_marker.line_number if position_marker else 0
        column_number = (
            position_marker.index_number
            + position_marker.index_indent
            + 1
            - column_delta
            if position_marker
            else 0
        )
        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_blank_line,
            extracted_whitespace,
            line_number=line_number,
            column_number=column_number,
            extracted_whitespace=extracted_whitespace,
        )

    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_blank_line

    # pylint: enable=protected-access

    def register_for_markdown_transform(
        self, registration_function: RegisterMarkdownTransformHandlersProtocol
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            BlankLineMarkdownToken, BlankLineMarkdownToken.__rehydrate_blank_line, None
        )

    @staticmethod
    def __rehydrate_blank_line(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the blank line from the token.
        """
        # if (
        #     self.context.block_stack
        #     and self.context.block_stack[-1].is_fenced_code_block
        #     and (previous_token and previous_token.is_text)
        # ):
        #     extra_newline_after_text_token = ParserHelper.newline_character
        # else:
        _ = previous_token, context
        extra_newline_after_text_token = ""

        current_blank_token = cast(BlankLineMarkdownToken, current_token)
        return f"{extra_newline_after_text_token}{current_blank_token.extracted_whitespace}{ParserHelper.newline_character}"

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            BlankLineMarkdownToken,
            BlankLineMarkdownToken.__handle_blank_line_token,
            None,
        )

    @staticmethod
    def __handle_blank_line_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = next_token

        if transform_state.is_in_html_block:
            output_html = f"{output_html}{ParserHelper.newline_character}"
        return output_html



import copy
import logging
from typing import List, Optional, Tuple, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import EndMarkdownToken, MarkdownToken
from pymarkdown.transform_markdown.markdown_transform_context import (
    IndentAdjustment,
    MarkdownTransformContext,
)

POGGER = ParserLogger(logging.getLogger(__name__))


class TransformBlockQuote:
    @staticmethod
    def rehydrate_block_quote(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
        next_token: Optional[MarkdownToken],
        transformed_data: str,
    ) -> str:
        _ = (previous_token, transformed_data)

        current_block_token = cast(BlockQuoteMarkdownToken, current_token)
        (
            token_stack_index,
            are_tokens_viable,
            new_instance,
        ) = TransformBlockQuote.__rehydrate_block_quote_start(
            context, current_block_token
        )

        if are_tokens_viable:
            container_list_token = cast(
                ListStartMarkdownToken, context.container_token_stack[token_stack_index]
            )
            matching_list_token = (
                container_list_token.last_new_list_token or container_list_token
            )
            POGGER.debug(
                f">matching_list_token>{ParserHelper.make_value_visible(matching_list_token)}"
            )

            POGGER.debug(f">current_token.line_number>{current_token.line_number}")
            POGGER.debug(
                ">container_token_stack[token_stack_index].line_number>"
                + f"{container_list_token.line_number}"
            )

        if (
            are_tokens_viable
            and current_token.line_number == matching_list_token.line_number
        ):
            container_list_token = cast(
                ListStartMarkdownToken, context.container_token_stack[token_stack_index]
            )
            already_existing_whitespace = ParserHelper.repeat_string(
                " ", container_list_token.indent_level
            )
        else:
            already_existing_whitespace = None

        POGGER.debug(
            f">bquote>current_token>{ParserHelper.make_value_visible(current_token)}"
        )
        POGGER.debug(
            f">bquote>next_token>{ParserHelper.make_value_visible(next_token)}"
        )

        selected_leading_sequence = (
            ""
            if next_token
            and next_token.is_block_quote_start
            and current_token.line_number == next_token.line_number
            else new_instance.calculate_next_bleading_space_part()
        )
        POGGER.debug(f">bquote>selected_leading_sequence>{selected_leading_sequence}<")

        POGGER.debug(
            f">bquote>already_existing_whitespace>:{already_existing_whitespace}:<"
        )
        POGGER.debug(
            f">bquote>selected_leading_sequence>:{selected_leading_sequence}:<"
        )
        if already_existing_whitespace and selected_leading_sequence.startswith(
            already_existing_whitespace
        ):
            selected_leading_sequence = selected_leading_sequence[
                len(already_existing_whitespace) :
            ]
            POGGER.debug(
                f">bquote>new selected_leading_sequence>{selected_leading_sequence}<"
            )
        return selected_leading_sequence

    @staticmethod
    def __rehydrate_block_quote_start(
        context: MarkdownTransformContext, current_token: BlockQuoteMarkdownToken
    ) -> Tuple[int, bool, BlockQuoteMarkdownToken]:
        POGGER.debug(
            f">bquote>tabbed_bleading_spaces>{ParserHelper.make_value_visible(current_token.tabbed_bleading_spaces)}"
        )
        new_instance = copy.deepcopy(current_token)
        POGGER.debug(
            f">bquote>tabbed_bleading_spaces>{ParserHelper.make_value_visible(new_instance.tabbed_bleading_spaces)}"
        )
        new_instance.leading_text_index = 0
        context.container_token_stack.append(new_instance)
        context.original_container_token_stack.append(current_token)
        context.container_token_indents.append(IndentAdjustment())

        POGGER.debug(f">bquote>{ParserHelper.make_value_visible(new_instance)}")
        POGGER.debug(
            f">self.container_token_stack>{ParserHelper.make_value_visible(context.container_token_stack)}"
        )
        token_stack_index = TransformBlockQuote.__search_backward_for_block_quote_start(
            context
        )
        are_tokens_viable = (
            len(context.container_token_stack) > 1 and token_stack_index >= 0
        )
        POGGER.debug(f">are_tokens_viable>{are_tokens_viable}")
        return token_stack_index, are_tokens_viable, new_instance

    @staticmethod
    def __search_backward_for_block_quote_start(
        context: MarkdownTransformContext,
    ) -> int:
        token_stack_index = len(context.container_token_stack) - 2
        while (
            token_stack_index >= 0
            and context.container_token_stack[token_stack_index].is_block_quote_start
        ):
            token_stack_index -= 1
        POGGER.debug(f">token_stack_index>{token_stack_index}")
        POGGER.debug(
            f">token_stack_token-->{ParserHelper.make_value_visible(context.container_token_stack[token_stack_index])}"
        )
        return token_stack_index

    @staticmethod
    def __rehydrate_block_quote_end_calc(
        current_token: MarkdownToken,
    ) -> Tuple[BlockQuoteMarkdownToken, Optional[str], int, List[str]]:
        current_end_token = cast(EndMarkdownToken, current_token)
        current_start_token = cast(
            BlockQuoteMarkdownToken, current_end_token.start_markdown_token
        )

        # POGGER.debug(
        #     f"current_start_block_token>:{ParserHelper.make_value_visible(current_start_token)}:<"
        # )
        current_end_token_extra = current_end_token.extra_end_data
        POGGER.debug(
            f"current_end_token_extra>:{ParserHelper.make_value_visible(current_end_token_extra)}:<"
        )
        start_leading_index = current_start_token.leading_text_index

        assert (
            current_start_token.bleading_spaces is not None
        ), "Bleading spaces should be defined by now."
        split_start_leading = current_start_token.bleading_spaces.split(
            ParserHelper.newline_character
        )
        POGGER.debug(
            f"start_leading_index>>:{ParserHelper.make_value_visible(start_leading_index)}:<"
        )
        POGGER.debug(
            f"split_start_leading>>:{ParserHelper.make_value_visible(split_start_leading)}:<"
        )

        return (
            current_start_token,
            current_end_token_extra,
            start_leading_index,
            split_start_leading,
        )

    @staticmethod
    def rehydrate_block_quote_end(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        actual_tokens: List[MarkdownToken],
        token_index: int,
    ) -> str:
        POGGER.debug(">>__rehydrate_block_quote_end")
        _ = current_token

        (
            current_start_token,
            current_end_token_extra,
            start_leading_index,
            split_start_leading,
        ) = TransformBlockQuote.__rehydrate_block_quote_end_calc(current_token)

        adjusted_end_string = (
            current_end_token_extra
            if start_leading_index + 1 < len(split_start_leading)
            and current_end_token_extra is not None
            else ""
        )
        POGGER.debug(
            f">>{ParserHelper.make_value_visible(actual_tokens[token_index:])}"
        )
        search_index = token_index + 1
        while (
            search_index < len(actual_tokens)
            and actual_tokens[search_index].is_container_end_token
        ):
            search_index += 1
        POGGER.debug(f">>{search_index}")
        any_non_container_end_tokens = search_index < len(actual_tokens)
        POGGER.debug(f">>{any_non_container_end_tokens}")

        del context.container_token_indents[-1]
        if context.container_token_indents and any_non_container_end_tokens:
            indent_adjust = (
                actual_tokens[search_index].line_number
                - current_start_token.line_number
            )
            if actual_tokens[token_index - 1].is_blank_line:
                indent_adjust -= 1

            for indent_index in range(len(context.container_token_indents) - 1, -1, -1):
                if context.container_token_stack[indent_index].is_block_quote_start:
                    context.container_token_indents[
                        indent_index
                    ].adjustment += indent_adjust
                    break
        del context.container_token_stack[-1]
        del context.original_container_token_stack[-1]

        TransformBlockQuote.__apply_kludge_two(context, current_start_token)
        return adjusted_end_string

    @staticmethod
    def __apply_kludge_two(
        context: MarkdownTransformContext, current_start_token: BlockQuoteMarkdownToken
    ) -> None:
        found_block_quote_token: Optional[BlockQuoteMarkdownToken] = None
        for stack_index in range(
            len(context.original_container_token_stack) - 1, -1, -1
        ):
            stack_token = context.original_container_token_stack[stack_index]
            if stack_token.is_block_quote_start:
                found_block_quote_token = cast(BlockQuoteMarkdownToken, stack_token)
                break
        if found_block_quote_token is not None:
            assert current_start_token.bleading_spaces is not None
            newline_count = ParserHelper.count_newlines_in_text(
                current_start_token.bleading_spaces
            )
            if found_block_quote_token.weird_kludge_two is None:
                found_block_quote_token.weird_kludge_two = newline_count
            else:
                found_block_quote_token.weird_kludge_two += newline_count



"""
Module to provide processing for the block quotes.
"""

import logging
from typing import List, Optional, Tuple, cast

from pymarkdown.block_quotes.block_quote_count_helper import BlockQuoteCountHelper
from pymarkdown.block_quotes.block_quote_data import BlockQuoteData
from pymarkdown.block_quotes.block_quote_non_fenced_helper import (
    BlockQuoteNonFencedHelper,
)
from pymarkdown.container_blocks.container_grab_bag import ContainerGrabBag
from pymarkdown.general.constants import Constants
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.parser_state import ParserState
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.general.requeue_line_info import RequeueLineInfo
from pymarkdown.leaf_blocks.leaf_block_processor import LeafBlockProcessor
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.stack_token import (
    BlockQuoteStackToken,
    FencedCodeBlockStackToken,
    ListStackToken,
    StackToken,
)

POGGER = ParserLogger(logging.getLogger(__name__))


class BlockQuoteProcessor:
    """
    Class to provide processing for the block quotes.
    """

    # pylint: disable=too-many-arguments
    @staticmethod
    def __adjust_lazy_handling(
        parser_state: ParserState,
        line_to_parse: str,
        extracted_whitespace: str,
        was_paragraph_continuation: bool,
        original_line: str,
        index_indent: int,
    ) -> Tuple[bool, bool]:
        if (
            parser_state.token_stack[-1].is_paragraph
            and not parser_state.token_document[-1].is_blank_line
        ):
            was_paragraph_continuation = True
            POGGER.debug("was_paragraph_continuation>>$", was_paragraph_continuation)

            is_leaf_block_start = (
                LeafBlockProcessor.is_paragraph_ending_leaf_block_start(
                    parser_state,
                    line_to_parse,
                    0,
                    extracted_whitespace,
                    original_line,
                    index_indent,
                )
            )

            POGGER.debug("is_leaf_block_start:$", is_leaf_block_start)
            if is_leaf_block_start:
                was_paragraph_continuation = False
                POGGER.debug(
                    "was_paragraph_continuation>>$", was_paragraph_continuation
                )
        else:
            is_leaf_block_start = False
        return was_paragraph_continuation, is_leaf_block_start

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def check_for_lazy_handling(
        parser_state: ParserState,
        position_marker: PositionMarker,
        block_quote_data: BlockQuoteData,
        line_to_parse: str,
        extracted_whitespace: str,
        was_paragraph_continuation: bool,
        original_line: str,
    ) -> Tuple[List[MarkdownToken], BlockQuoteData, bool]:
        """
        Check if there is any processing to be handled during the handling of
        lazy continuation lines in block quotes.
        """
        POGGER.debug("__check_for_lazy_handling")
        container_level_tokens: List[MarkdownToken] = []
        POGGER.debug(
            "block_quote_data.current_count>$>>block_quote_data.stack_count>>$<<",
            block_quote_data.current_count,
            block_quote_data.stack_count,
        )
        if block_quote_data.current_count == 0 and block_quote_data.stack_count > 0:
            POGGER.debug("haven't processed")

            POGGER.debug("xx:$", parser_state.token_stack)
            POGGER.debug("xx:$", parser_state.token_document)
            POGGER.debug("xx:$", parser_state.original_stack_depth)
            POGGER.debug("xx:$", position_marker.line_number)

            (
                was_paragraph_continuation,
                is_leaf_block_start,
            ) = BlockQuoteProcessor.__adjust_lazy_handling(
                parser_state,
                line_to_parse,
                extracted_whitespace,
                was_paragraph_continuation,
                original_line,
                position_marker.index_indent,
            )
            if (
                parser_state.token_stack[-1].is_code_block
                or parser_state.token_stack[-1].is_html_block
                or is_leaf_block_start
            ):
                POGGER.debug("__check_for_lazy_handling>>code block")
                container_level_tokens, _ = parser_state.close_open_blocks_fn(
                    parser_state,
                    only_these_blocks=[
                        BlockQuoteStackToken,
                        type(parser_state.token_stack[-1]),
                    ],
                    include_block_quotes=True,
                    was_forced=True,
                )
            stack_count = parser_state.count_of_block_quotes_on_stack()
            if stack_count != block_quote_data.stack_count:
                block_quote_data = BlockQuoteData(
                    block_quote_data.current_count, stack_count
                )
        return container_level_tokens, block_quote_data, was_paragraph_continuation

    # pylint: enable=too-many-arguments

    @staticmethod
    def __handle_block_quote_block_really_start(
        parser_state: ParserState,
        position_marker: PositionMarker,
        extracted_whitespace: str,
        grab_bag: ContainerGrabBag,
    ) -> Tuple[
        Optional[RequeueLineInfo],
        bool,
        int,
        bool,
        List[MarkdownToken],
        List[MarkdownToken],
    ]:
        assert (
            grab_bag.container_start_bq_count is not None
        ), "If starting here, we need a block quote count."
        POGGER.debug("handle_block_quote_block>>block-start")
        POGGER.debug("original_line:>:$:<", grab_bag.original_line)
        POGGER.debug(
            "container_start_bq_count:>:$:<", grab_bag.container_start_bq_count
        )
        (
            adjusted_text_to_parse,
            adjusted_index_number,
            leaf_tokens,
            container_level_tokens,
            grab_bag.block_quote_data,
            grab_bag.removed_chars_at_start_of_line,
            grab_bag.did_blank,
            last_block_quote_index,
            text_removed_by_container,
            avoid_block_starts,
            requeue_line_info,
            grab_bag.do_force_list_continuation,
        ) = BlockQuoteProcessor.__handle_block_quote_section(
            parser_state,
            position_marker,
            grab_bag.block_quote_data,
            extracted_whitespace,
            grab_bag.container_start_bq_count,
            grab_bag.original_line,
        )
        POGGER.debug("force_list_continuation=$", grab_bag.do_force_list_continuation)
        POGGER.debug("adjusted_index_number>>:$:", adjusted_index_number)
        POGGER.debug(">>avoid_block_starts>>$", avoid_block_starts)
        POGGER.debug(">>text_removed_by_container>>:$:", text_removed_by_container)

        (
            did_process,
            end_of_bquote_start_index,
        ) = BlockQuoteProcessor.__handle_block_quote_block_kludges(
            parser_state,
            grab_bag.block_quote_data,
            leaf_tokens,
            container_level_tokens,
            adjusted_text_to_parse,
            last_block_quote_index,
            adjusted_index_number,
        )
        grab_bag.start_index = adjusted_index_number
        grab_bag.line_to_parse = adjusted_text_to_parse
        grab_bag.text_removed_by_container = text_removed_by_container
        grab_bag.last_block_quote_index = last_block_quote_index
        return (
            requeue_line_info,
            did_process,
            end_of_bquote_start_index,
            avoid_block_starts,
            leaf_tokens,
            container_level_tokens,
        )

    @staticmethod
    def handle_block_quote_block(
        parser_state: ParserState,
        position_marker: PositionMarker,
        grab_bag: ContainerGrabBag,
    ) -> Tuple[bool, int, List[MarkdownToken], List[MarkdownToken], bool]:
        """
        Handle the processing of a block quote block.
        """
        POGGER.debug("handle_block_quote_block>>start")
        extracted_whitespace: Optional[str] = grab_bag.extracted_whitespace
        assert (
            extracted_whitespace is not None
        ), "Extracted whitespace must be available at this point."
        adj_ws: Optional[str] = grab_bag.adj_ws

        (
            did_process,
            avoid_block_starts,
            end_of_bquote_start_index,
        ) = (
            False,
            False,
            -1,
        )
        grab_bag.do_force_list_continuation = False
        grab_bag.did_blank = False
        grab_bag.removed_chars_at_start_of_line = 0
        grab_bag.start_index = position_marker.index_number
        grab_bag.line_to_parse = position_marker.text_to_parse
        grab_bag.text_removed_by_container = None
        grab_bag.last_block_quote_index = 0

        leaf_tokens: List[MarkdownToken] = []
        container_level_tokens: List[MarkdownToken] = []

        POGGER.debug(
            "handle_block_quote_block>>was_link_definition_started>:$:<",
            parser_state.token_stack[-1].was_link_definition_started,
        )
        POGGER.debug(
            "text_to_parse[index=$:]>:$:<",
            position_marker.index_number,
            position_marker.text_to_parse[position_marker.index_number :],
        )
        really_start, requeue_line_info = BlockQuoteProcessor.__check_if_really_start(
            parser_state, position_marker, extracted_whitespace, adj_ws
        )

        if really_start:
            assert (
                not requeue_line_info
            ), "If we need to start, we should not be requeuing a line."
            (
                requeue_line_info,
                did_process,
                end_of_bquote_start_index,
                avoid_block_starts,
                leaf_tokens,
                container_level_tokens,
            ) = BlockQuoteProcessor.__handle_block_quote_block_really_start(
                parser_state, position_marker, extracted_whitespace, grab_bag
            )
        elif (
            parser_state.token_stack[-1].was_link_definition_started
            and not requeue_line_info
        ):
            BlockQuoteProcessor.__handle_block_quote_block_lrd_kludges(parser_state)

        POGGER.debug("handle_block_quote_block>>end>>did_process>>$", did_process)
        POGGER.debug("handle_block_quote_block>>end>>leaf_tokens>>$", leaf_tokens)
        POGGER.debug(
            "handle_block_quote_block>>end>>container_level_tokens>>$",
            container_level_tokens,
        )

        grab_bag.requeue_line_info = requeue_line_info

        return (
            did_process,
            end_of_bquote_start_index,
            leaf_tokens,
            container_level_tokens,
            avoid_block_starts,
        )

    @staticmethod
    def __handle_block_quote_block_lrd_kludges(parser_state: ParserState) -> None:
        stack_index = parser_state.find_last_block_quote_on_stack()
        if stack_index > 0:
            last_block_token = cast(
                BlockQuoteMarkdownToken,
                parser_state.token_stack[stack_index].matching_markdown_token,
            )
            POGGER.debug(
                "handle_block w/ no open>>found>>$",
                last_block_token,
            )
            POGGER.debug("hbqblk>>last_block_token>>$", last_block_token)
            POGGER.debug(
                "hbqblk>>leading_text_index>>$", last_block_token.leading_text_index
            )
            last_block_token.add_bleading_spaces("")
            POGGER.debug("hbqblk>>last_block_token>>$", last_block_token)
            POGGER.debug(
                "hbqblk>>leading_text_index>>$", last_block_token.leading_text_index
            )

    # pylint: disable=too-many-arguments
    @staticmethod
    def __handle_block_quote_block_kludges(
        parser_state: ParserState,
        block_quote_data: BlockQuoteData,
        leaf_tokens: List[MarkdownToken],
        container_level_tokens: List[MarkdownToken],
        adjusted_text_to_parse: str,
        last_block_quote_index: int,
        adjusted_index_number: int,
    ) -> Tuple[bool, int]:
        POGGER.debug(
            ">>block_quote_data.current_count>>$", block_quote_data.current_count
        )
        POGGER.debug(">>block_quote_data.stack_count>>$", block_quote_data.stack_count)
        if block_quote_data.current_count:
            POGGER.debug("token_stack>$", parser_state.token_stack)
            POGGER.debug("token_document>$", parser_state.token_document)
            POGGER.debug("leaf_tokens>$", leaf_tokens)
            POGGER.debug("container_level_tokens>$", container_level_tokens)
            POGGER.debug("adjusted_text_to_parse>$<", adjusted_text_to_parse)
            adjusted_current_count = block_quote_data.current_count + 1
            if adjusted_current_count < len(parser_state.token_stack):
                POGGER.debug(
                    "token_stack[x]>$", parser_state.token_stack[adjusted_current_count]
                )
                if parser_state.token_stack[
                    adjusted_current_count
                ].is_list and adjusted_text_to_parse.strip(Constants.ascii_whitespace):
                    POGGER.debug("\n\nBOOM\n\n")
                    parser_state.nested_list_start = cast(
                        ListStackToken, parser_state.token_stack[adjusted_current_count]
                    )

        did_process = last_block_quote_index != -1
        return did_process, adjusted_index_number if did_process else -1

    # pylint: enable=too-many-arguments

    @staticmethod
    def __check_if_really_start(
        parser_state: ParserState,
        position_marker: PositionMarker,
        extracted_whitespace: str,
        adj_ws: Optional[str],
    ) -> Tuple[bool, Optional[RequeueLineInfo]]:
        POGGER.debug(
            "handle_block_quote_block>>text>:$:<", position_marker.text_to_parse
        )
        POGGER.debug(
            "handle_block_quote_block>>extracted_whitespace>:$:<",
            extracted_whitespace,
        )
        POGGER.debug("handle_block_quote_block>>adj_ws>:$:<", adj_ws)
        requeue_line_info = None

        last_stack_index = parser_state.find_last_list_block_on_stack()
        if (
            last_stack_index
            and len(extracted_whitespace) >= position_marker.index_number
        ):
            adj_ws = extracted_whitespace[position_marker.index_number - 1 :]
        POGGER.debug("handle_block_quote_block>>adj_ws>:$:<", adj_ws)
        really_start = BlockQuoteCountHelper.is_block_quote_start(
            position_marker.text_to_parse,
            position_marker.index_number,
            extracted_whitespace,
            adj_ws=adj_ws,
        )
        if (
            really_start
            and parser_state.token_stack[-1].is_fenced_code_block
            and parser_state.token_stack[-2].is_list
        ):
            start_index = position_marker.index_number
            POGGER.debug("start_index>>$", start_index)
            list_token = cast(
                ListStartMarkdownToken,
                parser_state.token_stack[-2].matching_markdown_token,
            )
            POGGER.debug("list_token>>$", list_token)
            POGGER.debug("list_token.indent_level>>$", list_token.indent_level)
            really_start = start_index < list_token.indent_level
            POGGER.debug("really_start>>$", really_start)

        if really_start:
            POGGER.debug(
                "handle_block_quote_block>>token_stack[depth]>:$:<",
                parser_state.token_stack,
            )
            if parser_state.token_stack[-1].is_paragraph:
                (
                    really_start,
                    requeue_line_info,
                ) = BlockQuoteProcessor.__check_if_really_start_paragraph(
                    parser_state, position_marker
                )
        return really_start, requeue_line_info

    @staticmethod
    def __check_if_really_start_paragraph(
        parser_state: ParserState, position_marker: PositionMarker
    ) -> Tuple[bool, Optional[RequeueLineInfo]]:
        current_indent, eligible_stack, eligible_stack_index = (
            0,
            parser_state.token_stack[1:-1],
            0,
        )

        POGGER.debug("handle_block_quote_block>>eligible_stack>:$:<", eligible_stack)
        while eligible_stack_index < len(eligible_stack):
            if eligible_stack[eligible_stack_index].is_list:
                (
                    current_indent,
                    requeue_line_info,
                ) = BlockQuoteProcessor.__check_if_really_start_list(
                    parser_state, position_marker, eligible_stack, eligible_stack_index
                )
                if requeue_line_info:
                    return False, requeue_line_info
            else:
                assert eligible_stack[
                    eligible_stack_index
                ].is_block_quote, (
                    "If its not a list token, it must be a block quote token."
                )
                break  # pragma: no cover
            eligible_stack_index += 1
        POGGER.debug(
            "eligible_stack_index($) < len(eligible_stack)($)",
            eligible_stack_index,
            len(eligible_stack),
        )
        POGGER.debug("current_indent($)", current_indent)
        return True, None

    @staticmethod
    def __check_if_really_start_list(
        parser_state: ParserState,
        position_marker: PositionMarker,
        eligible_stack: List[StackToken],
        eligible_stack_index: int,
    ) -> Tuple[int, Optional[RequeueLineInfo]]:
        assert eligible_stack[
            eligible_stack_index
        ].is_list, "Stack for lists should always end with list tokens."
        list_token = cast(ListStackToken, eligible_stack[eligible_stack_index])
        current_indent = list_token.indent_level
        if current_indent <= position_marker.index_number:
            return current_indent, None
        POGGER.debug("BOOYAH")
        POGGER.debug("current_indent=$", current_indent)
        POGGER.debug("index_number=$", position_marker.index_number)
        POGGER.debug("eligible_stack=$", eligible_stack)
        POGGER.debug("eligible_stack_index=$", eligible_stack_index)

        while eligible_stack_index >= 0:
            assert eligible_stack[
                eligible_stack_index
            ].is_list, "Stack for lists should always end with list tokens."
            list_token = cast(ListStackToken, eligible_stack[eligible_stack_index])
            if list_token.indent_level <= position_marker.index_number:
                break
            eligible_stack_index -= 1
        POGGER.debug("eligible_stack_index=$", eligible_stack_index)
        root_index = (
            (parser_state.token_stack.index(eligible_stack[eligible_stack_index]) + 1)
            if eligible_stack_index >= 0
            else 0
        )
        POGGER.debug("root_index=$", root_index)
        (
            container_level_tokens,
            _,
        ) = parser_state.close_open_blocks_fn(
            parser_state,
            include_block_quotes=True,
            include_lists=True,
            until_this_index=root_index,
            was_forced=True,
        )
        parser_state.token_document.extend(container_level_tokens)
        return -1, RequeueLineInfo([position_marker.text_to_parse], False, False)

    # pylint: disable=too-many-arguments
    @staticmethod
    def __handle_block_quote_section(
        parser_state: ParserState,
        position_marker: PositionMarker,
        block_quote_data: BlockQuoteData,
        extracted_whitespace: str,
        container_start_bq_count: int,
        original_line: str,
    ) -> Tuple[
        str,
        int,
        List[MarkdownToken],
        List[MarkdownToken],
        BlockQuoteData,
        int,
        bool,
        int,
        Optional[str],
        bool,
        Optional[RequeueLineInfo],
        bool,
    ]:
        """
        Handle the processing of a section clearly identified as having block quotes.
        """
        leaf_tokens: List[MarkdownToken] = []
        container_level_tokens: List[MarkdownToken] = []

        POGGER.debug(
            "IN>__handle_block_quote_section---$<<<",
            position_marker.text_to_parse,
        )
        POGGER.debug(
            "IN>start_index---$<<<",
            position_marker.index_number,
        )

        POGGER.debug("block_quote_data.stack_count--$", block_quote_data.stack_count)
        POGGER.debug("token_stack[-1]--$", parser_state.token_stack[-1])
        POGGER.debug("token_stack--$", parser_state.token_stack)

        POGGER.debug(
            "__handle_block_quote_section---$--$--",
            position_marker.index_number,
            position_marker.text_to_parse,
        )

        (
            block_quote_data,
            start_index,
            line_to_parse,
            last_block_quote_index,
            avoid_block_starts,
        ) = BlockQuoteCountHelper.count_block_quote_starts(
            parser_state,
            position_marker.text_to_parse,
            position_marker.index_number,
            block_quote_data,
            parser_state.token_stack[-1].is_fenced_code_block,
            parser_state.token_stack[-1].is_html_block,
        )
        POGGER.debug(
            "block_quote_data>>:curr=$:stack=$:",
            block_quote_data.current_count,
            block_quote_data.stack_count,
        )
        POGGER.debug("start_index>>:$:", start_index)
        POGGER.debug("line_to_parse>>:$:", line_to_parse)
        POGGER.debug("last_block_quote_index>>:$:", last_block_quote_index)
        POGGER.debug(">>avoid_block_starts>>$", avoid_block_starts)

        POGGER.debug("token_stack--$", parser_state.token_stack)
        POGGER.debug(">>container_start_bq_count>>$", container_start_bq_count)
        POGGER.debug(">>original_start_index>>$", position_marker.index_number)

        if last_block_quote_index != -1:
            POGGER.debug("start_index>>:$:", start_index)
            return BlockQuoteProcessor.__handle_existing_block_quote(
                parser_state,
                block_quote_data,
                start_index,
                line_to_parse,
                extracted_whitespace,
                position_marker.index_number,
                container_start_bq_count,
                position_marker,
                leaf_tokens,
                container_level_tokens,
                original_line,
                last_block_quote_index,
                avoid_block_starts,
            )

        return (
            line_to_parse,
            start_index,
            leaf_tokens,
            container_level_tokens,
            block_quote_data,
            0,
            False,
            last_block_quote_index,
            None,
            avoid_block_starts,
            None,
            False,
        )

    # pylint: enable=too-many-arguments

    @staticmethod
    def __handle_existing_block_quote_fenced_special(
        parser_state: ParserState, start_index: int, block_quote_data: BlockQuoteData
    ) -> Tuple[bool, int]:
        assert parser_state.original_line_to_parse is not None
        block_quote_character_count = ParserHelper.count_characters_in_text(
            parser_state.original_line_to_parse[:start_index], ">"
        )
        assert (
            block_quote_character_count <= block_quote_data.current_count
        ), "if not, overreach"
        count_block_quotes = 0
        find_token_index = 0
        for stack_token_index, stack_token in enumerate(
            parser_state.token_stack
        ):  # pragma: no cover
            if stack_token.is_block_quote:
                count_block_quotes += 1
                if count_block_quotes == block_quote_character_count:
                    find_token_index = stack_token_index
                    break
        assert find_token_index != len(
            parser_state.token_stack
        ), "should have completed before this"
        find_token_index += 1
        process_fenced_block = parser_state.token_stack[find_token_index].is_block_quote
        return process_fenced_block, find_token_index

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def __handle_existing_block_quote_fenced_special_part_two(
        parser_state: ParserState,
        stack_index: int,
        line_to_parse: str,
        start_index: int,
        block_quote_data: BlockQuoteData,
        leaf_tokens: List[MarkdownToken],
        container_level_tokens: List[MarkdownToken],
        avoid_block_starts: bool,
    ) -> Tuple[
        str,
        int,
        List[MarkdownToken],
        List[MarkdownToken],
        BlockQuoteData,
        int,
        bool,
        int,
        Optional[str],
        bool,
        Optional[RequeueLineInfo],
        bool,
    ]:
        # At this point, we have a "prefix", which may be partial, that has the
        # current_count of > characters, and ends with a list. If we are here,
        # we know that previous lines have had at least one more > character and
        # counted block quote.
        assert parser_state.token_stack[
            stack_index
        ].is_list, "If not bq, must be a list."
        while parser_state.token_stack[stack_index].is_list:
            stack_index += 1
        embedded_list_stack_token = cast(
            ListStackToken, parser_state.token_stack[stack_index - 1]
        )
        block_stack_token = parser_state.token_stack[stack_index]
        block_markdown_token = cast(
            BlockQuoteMarkdownToken, block_stack_token.matching_markdown_token
        )
        assert parser_state.original_line_to_parse is not None
        if character_after_list := parser_state.original_line_to_parse[
            start_index : embedded_list_stack_token.indent_level
        ].strip():
            raise ValueError(
                f"probably list start character ({character_after_list}) encountered"
            )
        assert block_quote_data.current_count + 1 == block_quote_data.stack_count
        sd = parser_state.original_line_to_parse[embedded_list_stack_token.indent_level]
        assert sd == ">"
        last_block_quote_index = embedded_list_stack_token.indent_level + 1
        character_after_block_quote = parser_state.original_line_to_parse[
            last_block_quote_index
        ]
        assert character_after_block_quote == " "
        # if character_after_block_quote == " ":
        last_block_quote_index += 1
        character_after_block_quote = parser_state.original_line_to_parse[
            last_block_quote_index
        ]

        text_removed_by_container = parser_state.original_line_to_parse[
            :last_block_quote_index
        ]
        POGGER.debug(
            "__handle_existing_block_quote_fenced_special_part_two>>block_token>>$",
            block_markdown_token,
        )
        block_markdown_token.add_bleading_spaces(text_removed_by_container)
        POGGER.debug(
            "__handle_existing_block_quote_fenced_special_part_two>>block_token>>$",
            block_markdown_token,
        )
        if block_markdown_token.weird_kludge_one:
            block_markdown_token.weird_kludge_one += 1
        else:
            block_markdown_token.weird_kludge_one = 1
        block_quote_data = BlockQuoteData(
            block_quote_data.current_count + 1, block_quote_data.stack_count
        )
        return (
            line_to_parse[last_block_quote_index:],
            last_block_quote_index,
            leaf_tokens,
            container_level_tokens,
            block_quote_data,
            0,
            False,
            last_block_quote_index,
            text_removed_by_container,
            avoid_block_starts,
            None,
            False,
        )

    # pylint: enable=too-many-arguments, too-many-locals

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def __handle_existing_block_quote(
        parser_state: ParserState,
        block_quote_data: BlockQuoteData,
        start_index: int,
        line_to_parse: str,
        extracted_whitespace: str,
        original_start_index: int,
        container_start_bq_count: int,
        position_marker: PositionMarker,
        leaf_tokens: List[MarkdownToken],
        container_level_tokens: List[MarkdownToken],
        original_line: str,
        last_block_quote_index: int,
        avoid_block_starts: bool,
    ) -> Tuple[
        str,
        int,
        List[MarkdownToken],
        List[MarkdownToken],
        BlockQuoteData,
        int,
        bool,
        int,
        Optional[str],
        bool,
        Optional[RequeueLineInfo],
        bool,
    ]:
        POGGER.debug(
            "__handle_block_quote_section---block_quote_data.current_count--$--$--$--",
            block_quote_data.current_count,
            start_index,
            line_to_parse,
        )
        POGGER.debug(
            "ORIG-->WS[$]--SI[$]--[$]",
            extracted_whitespace,
            original_start_index,
            parser_state.original_line_to_parse,
        )
        POGGER.debug("NOW -->SI[$]--[$]", start_index, line_to_parse)

        if container_start_bq_count:
            POGGER.debug(
                ">>block_quote_data.current_count>>$", block_quote_data.current_count
            )
            block_quote_data = BlockQuoteData(
                block_quote_data.current_count + container_start_bq_count,
                block_quote_data.stack_count,
            )

        process_fenced_block = parser_state.token_stack[-1].is_fenced_code_block
        if (
            process_fenced_block
            and block_quote_data.current_count < block_quote_data.stack_count
        ):
            xprocess_fenced_block, xstack_index = (
                BlockQuoteProcessor.__handle_existing_block_quote_fenced_special(
                    parser_state, start_index, block_quote_data
                )
            )
            if not xprocess_fenced_block:
                try:
                    return BlockQuoteProcessor.__handle_existing_block_quote_fenced_special_part_two(
                        parser_state,
                        xstack_index,
                        line_to_parse,
                        start_index,
                        block_quote_data,
                        leaf_tokens,
                        container_level_tokens,
                        avoid_block_starts,
                    )
                except ValueError as this_exception:
                    POGGER.debug(f"escaped special handling due to: {this_exception}")

        return BlockQuoteProcessor.__handlers(
            process_fenced_block,
            parser_state,
            block_quote_data,
            extracted_whitespace,
            position_marker,
            original_start_index,
            container_start_bq_count,
            line_to_parse,
            start_index,
            leaf_tokens,
            original_line,
            last_block_quote_index,
            avoid_block_starts,
            container_level_tokens,
        )

    # pylint: enable=too-many-arguments, too-many-locals

    # pylint: disable=too-many-arguments
    @staticmethod
    def __handlers(
        process_fenced_block: bool,
        parser_state: ParserState,
        block_quote_data: BlockQuoteData,
        extracted_whitespace: str,
        position_marker: PositionMarker,
        original_start_index: int,
        container_start_bq_count: int,
        line_to_parse: str,
        start_index: int,
        leaf_tokens: List[MarkdownToken],
        original_line: str,
        last_block_quote_index: int,
        avoid_block_starts: bool,
        container_level_tokens: List[MarkdownToken],
    ) -> Tuple[
        str,
        int,
        List[MarkdownToken],
        List[MarkdownToken],
        BlockQuoteData,
        int,
        bool,
        int,
        Optional[str],
        bool,
        Optional[RequeueLineInfo],
        bool,
    ]:
        if not process_fenced_block:
            return BlockQuoteNonFencedHelper.handle_non_fenced_code_section(
                parser_state,
                block_quote_data,
                extracted_whitespace,
                position_marker,
                original_start_index,
                container_start_bq_count,
                line_to_parse,
                start_index,
                leaf_tokens,
                original_line,
                last_block_quote_index,
                avoid_block_starts,
            )

        (
            block_quote_data,
            line_to_parse,
            container_level_tokens,
            text_removed_by_container,
        ) = BlockQuoteProcessor.__handle_fenced_code_section(
            parser_state,
            block_quote_data,
            start_index,
            line_to_parse,
            container_level_tokens,
        )

        POGGER.debug(
            "OUT>__handle_block_quote_section---$<<<",
            line_to_parse,
        )
        POGGER.debug("force_list_continuation=false")
        POGGER.debug("start_index>>:$:", start_index)
        return (
            line_to_parse,
            start_index,
            leaf_tokens,
            container_level_tokens,
            block_quote_data,
            0,
            False,
            last_block_quote_index,
            text_removed_by_container,
            avoid_block_starts,
            None,
            False,
        )

    # pylint: enable=too-many-arguments

    @staticmethod
    def __handle_fenced_code_section(
        parser_state: ParserState,
        block_quote_data: BlockQuoteData,
        start_index: int,
        line_to_parse: str,
        container_level_tokens: List[MarkdownToken],
    ) -> Tuple[BlockQuoteData, str, List[MarkdownToken], str]:
        POGGER.debug("handle_block_quote_section>>fenced")
        assert start_index >= 0, "Index must be valid."
        removed_text, line_to_parse = (
            line_to_parse[:start_index],
            line_to_parse[start_index:],
        )

        POGGER.debug("__hbqs>>removed_text>>$", removed_text)
        POGGER.debug("__hbqs>>line_to_parse>>$", line_to_parse)
        POGGER.debug(
            "__hbqs>>block_quote_data.current_count>>$", block_quote_data.current_count
        )
        POGGER.debug(
            "__hbqs>>block_quote_data.stack_count>>$", block_quote_data.stack_count
        )

        if block_quote_data.current_count < block_quote_data.stack_count:
            (
                container_level_tokens,
                _,
            ) = parser_state.close_open_blocks_fn(
                parser_state,
                only_these_blocks=[
                    FencedCodeBlockStackToken,
                ],
                was_forced=True,
            )
            block_quote_data = BlockQuoteCountHelper.decrease_stack_to_level(
                parser_state,
                block_quote_data.current_count,
                block_quote_data.stack_count,
                container_level_tokens,
            )

        stack_index = parser_state.find_last_block_quote_on_stack()
        found_bq_stack_token = parser_state.token_stack[stack_index]
        POGGER.debug(
            "found_bq_stack_token---$<<<",
            found_bq_stack_token,
        )
        found_bq_token = cast(
            BlockQuoteMarkdownToken, found_bq_stack_token.matching_markdown_token
        )
        POGGER.debug("hfcs>>last_block_token>>$", found_bq_token)
        POGGER.debug("hfcs>>leading_text_index>>$", found_bq_token.leading_text_index)
        found_bq_token.add_bleading_spaces(removed_text)
        found_bq_token.leading_text_index += 1
        POGGER.debug("hfcs>>last_block_token>>$", found_bq_token)
        POGGER.debug("hfcs>>leading_text_index>>$", found_bq_token.leading_text_index)
        text_removed_by_container = removed_text

        return (
            block_quote_data,
            line_to_parse,
            container_level_tokens,
            text_removed_by_container,
        )



"""
Module to provide for a simple implementation of a title case algorithm.
"""

import argparse
import logging
import os
import runpy
import traceback
from typing import List, Optional, Tuple, cast

from application_properties import ApplicationProperties
from application_properties.application_properties_utilities import (
    ApplicationPropertiesUtilities,
)

from pymarkdown.application_configuration_helper import ApplicationConfigurationHelper
from pymarkdown.application_file_scanner import (
    ApplicationFileScanner,
    ApplicationFileScannerOutputProtocol,
)
from pymarkdown.application_logging import ApplicationLogging
from pymarkdown.extension_manager.extension_manager import ExtensionManager
from pymarkdown.file_scan_helper import FileScanHelper
from pymarkdown.general.bad_tokenization_error import BadTokenizationError
from pymarkdown.general.main_presentation import MainPresentation
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.tokenized_markdown import TokenizedMarkdown
from pymarkdown.plugin_manager.bad_plugin_error import BadPluginError
from pymarkdown.plugin_manager.plugin_manager import PluginManager
from pymarkdown.return_code_helper import ApplicationResult, ReturnCodeHelper

POGGER = ParserLogger(logging.getLogger(__name__))

LOGGER = logging.getLogger(__name__)


# pylint: disable=too-many-instance-attributes
class PyMarkdownLint:
    """
    Class to provide for a simple implementation of a title case algorithm.
    """

    def __init__(
        self,
        presentation: Optional[MainPresentation] = None,
        show_stack_trace: bool = False,
        inherit_logging: bool = False,
        string_to_scan: Optional[str] = None,
    ) -> None:
        self.__presentation = presentation or MainPresentation()
        self.__show_stack_trace = show_stack_trace
        self.__string_to_scan = string_to_scan

        self.__version_number = PyMarkdownLint.__get_semantic_version()
        self.__properties: ApplicationProperties = ApplicationProperties()
        self.__logging = (
            None
            if inherit_logging
            else ApplicationLogging(
                self.__properties,
                default_log_level="CRITICAL",
                show_stack_trace=show_stack_trace,
            )
        )
        self.__tokenizer: Optional[TokenizedMarkdown] = None
        self.__plugins: PluginManager = PluginManager(self.__presentation)
        self.__extensions: ExtensionManager = ExtensionManager(self.__presentation)

    @property
    def application_version(self) -> str:
        """
        Return the version of the application.
        """
        return self.__version_number

    @staticmethod
    def __get_semantic_version() -> str:
        file_path = __file__
        assert os.path.isabs(
            file_path
        ), "This should hold on all operating systems.  This is to be sure of that assumption."
        file_path = file_path.replace(os.sep, "/")
        last_index = file_path.rindex("/")
        file_path = f"{file_path[: last_index + 1]}version.py"
        version_meta = runpy.run_path(file_path)
        return str(version_meta["__version__"])

    def __initialize_subsystems(
        self, direct_args: Optional[List[str]]
    ) -> argparse.Namespace:
        ReturnCodeHelper.reset()

        args = self.__parse_arguments(direct_args=direct_args)
        self.__set_initial_state(args)

        self.__show_stack_trace = args.show_stack_trace
        if not self.__show_stack_trace and self.__properties:
            self.__show_stack_trace = (
                self.__properties.get_boolean_property("log.stack-trace") or False
            )

        if self.__logging:
            self.__logging.initialize(args)
        ParserLogger.sync_on_next_call()

        if direct_args is None:
            LOGGER.debug("Using supplied command line arguments.")
        else:
            LOGGER.debug("Using direct arguments: %s", str(direct_args))

        self.__initialize_plugins_and_extensions(args)
        return args

    def __parse_arguments(self, direct_args: Optional[List[str]]) -> argparse.Namespace:
        parser = argparse.ArgumentParser(description="Lint any found Markdown files.")

        parser.add_argument(
            "-e",
            "--enable-rules",
            dest="enable_rules",
            action="store",
            default="",
            help="comma separated list of rules to enable",
        )
        parser.add_argument(
            "-d",
            "--disable-rules",
            dest="disable_rules",
            action="store",
            default="",
            help="comma separated list of rules to disable",
        )
        parser.add_argument(
            "--enable-extensions",
            dest="enable_extensions",
            action="store",
            default="",
            help="comma separated list of extensions to enable",
        )

        parser.add_argument(
            "-x-stdin",
            dest="x_test_stdin_fault",
            action="store_true",
            default="",
            help=argparse.SUPPRESS,
        )
        parser.add_argument(
            "-x-fix-debug",
            dest="x_fix_debug",
            action="store_true",
            default="",
            help=argparse.SUPPRESS,
        )
        parser.add_argument(
            "-x-fix-file-debug",
            dest="x_fix_file_debug",
            action="store_true",
            default="",
            help=argparse.SUPPRESS,
        )
        parser.add_argument(
            "-x-fix-no-rescan-log",
            dest="x_fix_no_rescan_log",
            action="store_true",
            default="",
            help=argparse.SUPPRESS,
        )

        parser.add_argument(
            "--add-plugin",
            dest="add_plugin",
            action="append",
            default=None,
            help="path to a plugin containing a new rule to apply",
        )
        ApplicationPropertiesUtilities.add_default_command_line_arguments(parser)
        parser.add_argument(
            "--no-json5",
            dest="use_json5_for_configuration",
            action="store_false",
            default=True,
            help="use stdlib's json reader instead of new JSON5 json reader",
        )
        parser.add_argument(
            "--stack-trace",
            dest="show_stack_trace",
            action="store_true",
            default=False,
            help="if an error occurs, print out the stack trace for debug purposes",
        )
        parser.add_argument(
            "--continue-on-error",
            dest="continue_on_error",
            action="store_true",
            default=False,
            help="if a tokenization or plugin error occurs, allow processing to continue",
        )
        ApplicationLogging.add_default_command_line_arguments(parser)
        ReturnCodeHelper.add_command_line_arguments(parser)

        subparsers = parser.add_subparsers(dest="primary_subparser")
        ExtensionManager.add_argparse_subparser(subparsers)
        FileScanHelper.add_argparse_subparser(subparsers, True)
        PluginManager.add_argparse_subparser(subparsers)
        FileScanHelper.add_argparse_subparser(subparsers, False)

        subparsers.add_parser("version", help="version of the application")

        parse_arguments = parser.parse_args(args=direct_args)

        if not parse_arguments.primary_subparser:
            parser.print_help()
            ReturnCodeHelper.exit_application(ApplicationResult.COMMAND_LINE_ERROR)
        elif parse_arguments.primary_subparser == "version":
            print(f"{self.__version_number}")
            ReturnCodeHelper.exit_application(ApplicationResult.SUCCESS)
        return parse_arguments

    def __set_initial_state(self, args: argparse.Namespace) -> None:
        if self.__logging:
            self.__logging.pre_initialize_with_args(args)

        ApplicationConfigurationHelper.apply_configuration_layers(
            args, self.__properties, self.__handle_error
        )

        ReturnCodeHelper.set_initial_state(args, self.__properties)

        LOGGER.info("Configuration loaded and applied.  Initial state setup completed.")

    def __initialize_plugins_and_extensions(self, args: argparse.Namespace) -> None:
        self.__initialize_plugins(args)
        self.__initialize_extensions(args)

        if args.primary_subparser == PluginManager.argparse_subparser_name():
            ReturnCodeHelper.exit_application(
                self.__plugins.handle_argparse_subparser(args)
            )
        if args.primary_subparser == ExtensionManager.argparse_subparser_name():
            ReturnCodeHelper.exit_application(
                self.__extensions.handle_argparse_subparser(args)
            )

    def __initialize_plugins(self, args: argparse.Namespace) -> None:
        try:
            plugin_dir = os.path.join(
                os.path.dirname(os.path.realpath(__file__)), "plugins"
            )
            self.__initialize_plugin_manager(args, plugin_dir)
            self.__apply_configuration_to_plugins()
        except ValueError as this_exception:
            self.__handle_error(
                f"{type(this_exception).__name__} encountered while initializing plugins:\n"
                + f"{this_exception}",
                this_exception,
            )

    def __initialize_plugin_manager(
        self, args: argparse.Namespace, plugin_dir: str
    ) -> None:
        """
        Make sure all plugins are ready before being initialized.
        """

        try:
            self.__plugins.initialize(
                plugin_dir,
                args.add_plugin,
                args.enable_rules,
                args.disable_rules,
                self.__properties,
                self.__show_stack_trace,
                args.x_fix_debug,
            )
        except BadPluginError as this_exception:
            self.__handle_error(
                f"BadPluginError encountered while loading plugins:\n{this_exception}",
                this_exception,
            )

    # pylint: disable=broad-exception-caught
    def __apply_configuration_to_plugins(self) -> None:
        try:
            self.__plugins.apply_configuration(self.__properties)
        except Exception as this_exception:
            self.__handle_error(
                f"{type(this_exception).__name__} encountered while configuring plugins:\n"
                + f"{this_exception}",
                this_exception,
            )

    # pylint: enable=broad-exception-caught

    # pylint: disable=broad-exception-caught
    def __initialize_extensions(self, args: argparse.Namespace) -> None:
        try:
            self.__extensions.initialize(
                self.__properties,
            )
            self.__extensions.apply_configuration(args.enable_extensions)

        except ValueError as this_exception:
            formatted_error = (
                f"Configuration error {type(this_exception).__name__} encountered "
                + f"while initializing extensions:\n{this_exception}"
            )
            self.__handle_error(formatted_error, this_exception)
        except Exception as this_exception:
            formatted_error = (
                f"Error {type(this_exception).__name__} encountered while initializing extensions:\n"
                + f"{this_exception}"
            )
            self.__handle_error(formatted_error, this_exception)

    # pylint: enable=broad-exception-caught

    def __initialize_parser(self) -> None:
        try:
            self.__tokenizer = TokenizedMarkdown()
            self.__tokenizer.apply_configuration(self.__properties, self.__extensions)
        except BadTokenizationError as this_exception:
            formatted_error = (
                f"{type(this_exception).__name__} encountered while initializing tokenizer:\n"
                + f"{this_exception}"
            )
            self.__handle_error(formatted_error, this_exception)

    def __handle_error(
        self,
        formatted_error: str,
        thrown_error: Optional[Exception],
        exit_on_error: bool = True,
        print_prefix: str = "\n\n",
    ) -> None:
        LOGGER.warning(formatted_error, exc_info=thrown_error)

        stack_trace = (
            "\n" + traceback.format_exc()
            if self.__show_stack_trace
            and thrown_error
            and not isinstance(thrown_error, ValueError)
            else ""
        )
        self.__presentation.print_system_error(
            f"{print_prefix}{formatted_error}{stack_trace}"
        )
        if exit_on_error:
            ReturnCodeHelper.exit_application(ApplicationResult.SYSTEM_ERROR)

    def __handle_file_scanner_output(self, formatted_output: str) -> None:
        self.__presentation.print_system_output(formatted_output)

    def __handle_file_scanner_error(self, formatted_error: str) -> None:
        self.__handle_error(formatted_error, None, exit_on_error=False)

    def __scan_files_if_no_errors(
        self,
        args: argparse.Namespace,
        use_standard_in: bool,
        files_to_scan: List[str],
        did_error_scanning_files: bool,
    ) -> ApplicationResult:  # sourcery skip: extract-method
        did_fix_any_files = False
        if did_error_scanning_files:
            self.__handle_error("No matching files found.", None, exit_on_error=False)
            return ApplicationResult.NO_FILES_TO_SCAN

        POGGER.info("Initializing parser.")
        self.__initialize_parser()

        POGGER.info("Processing files with parser.")
        assert (
            self.__tokenizer is not None
        ), "When scanning, the tokenizer should have already been initialized."
        fsh = FileScanHelper(
            self.__tokenizer,
            self.__plugins,
            self.__presentation,
            self.__show_stack_trace,
            self.__handle_error,
        )
        did_fix_any_files, did_fail_any_file = fsh.process_files_to_scan(
            args, use_standard_in, files_to_scan, self.__string_to_scan
        )
        if did_fail_any_file:
            scan_result = ApplicationResult.SYSTEM_ERROR
        elif did_fix_any_files:
            scan_result = ApplicationResult.FIXED_AT_LEAST_ONE_FILE
        elif self.__plugins.number_of_scan_failures:
            scan_result = ApplicationResult.SCAN_TRIGGERED_AT_LEAST_ONCE
        else:
            scan_result = ApplicationResult.SUCCESS

        POGGER.info("Files have been processed.")
        return scan_result

    def __find_files_to_scan(
        self, args: argparse.Namespace
    ) -> Tuple[bool, List[str], bool, bool]:
        use_standard_in = FileScanHelper.is_scan_stdin_specified(args)
        if use_standard_in:
            return use_standard_in, [], False, False

        POGGER.info("Determining files to scan.")
        paths_to_exclude: List[str] = args.path_exclusions or []
        if config_paths := self.__properties.get_string_property("system.exclude_path"):
            paths_to_exclude.extend(config_paths.split(","))

        files_to_scan, did_error_scanning_files, did_only_list_files = (
            ApplicationFileScanner.determine_files_to_scan_with_args(
                args,
                cast(
                    ApplicationFileScannerOutputProtocol,
                    self.__handle_file_scanner_output,
                ),
                cast(
                    ApplicationFileScannerOutputProtocol,
                    self.__handle_file_scanner_error,
                ),
                paths_to_exclude,
            )
        )
        return (
            use_standard_in,
            files_to_scan,
            did_error_scanning_files,
            did_only_list_files,
        )

    # pylint: disable=broad-exception-caught
    def main(self, direct_args: Optional[List[str]] = None) -> None:
        """
        Main entrance point.
        """
        try:
            args = self.__initialize_subsystems(direct_args)

            (
                use_standard_in,
                files_to_scan,
                did_error_scanning_files,
                did_only_list_files,
            ) = self.__find_files_to_scan(args)
            if did_only_list_files:
                scan_result = (
                    ApplicationResult.SUCCESS
                    if files_to_scan
                    else ApplicationResult.NO_FILES_TO_SCAN
                )
            else:
                scan_result = self.__scan_files_if_no_errors(
                    args, use_standard_in, files_to_scan, did_error_scanning_files
                )
            ReturnCodeHelper.exit_application(scan_result)
        except ValueError as this_exception:
            formatted_error = f"Configuration Error: {this_exception}"
            self.__handle_error(formatted_error, this_exception)
        except Exception as this_exception:
            formatted_error = (
                f"Unexpected Error({type(this_exception).__name__}): {this_exception}"
            )
            self.__handle_error(formatted_error, this_exception)
        finally:
            if self.__logging:
                self.__logging.terminate()
            self.__logging = None

    # pylint: enable=broad-exception-caught


# pylint: enable=too-many-instance-attributes


if __name__ == "__main__":
    PyMarkdownLint().main()



"""
Module to implement a plugin that looks for single line emphasis text that looks
like it is being used instead of a heading.
"""

from enum import Enum
from typing import List, Optional, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.plugin_manager.plugin_details import (
    PluginDetails,
    PluginDetailsV3,
    QueryConfigItem,
)
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken


class RuleMd036States(Enum):
    """
    Enumeration to provide guidance on what to look for as the tokens come in.
    """

    LOOK_FOR_PARAGRAPH = 0
    LOOK_FOR_EMPHASIS_START = 1
    LOOK_FOR_ELIGIBLE_TEXT = 2
    LOOK_FOR_EMPHASIS_END = 3
    LOOK_FOR_PARAGRAPH_END = 4


class RuleMd036(RulePlugin):
    """
    Class to implement a plugin that looks for single line emphasis text that looks
    like it is being used instead of a heading.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__punctuation: str = ""
        self.__current_state: RuleMd036States = RuleMd036States.LOOK_FOR_PARAGRAPH
        self.__start_token: Optional[MarkdownToken] = None

    def get_details(self) -> PluginDetails:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV3(
            plugin_name="no-emphasis-as-heading,no-emphasis-as-header",
            plugin_id="MD036",
            plugin_enabled_by_default=True,
            plugin_description="Emphasis possibly used instead of a heading element.",
            plugin_version="0.6.0",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md036.md",
            plugin_configuration="punctuation",
        )

    def initialize_from_config(self) -> None:
        """
        Event to allow the plugin to load configuration information.
        """
        self.__punctuation = self.plugin_configuration.get_string_property_with_default(
            "punctuation", ".,;:!?。，；：？"
        )

    def query_config(self) -> List[QueryConfigItem]:
        """
        Query to find out the configuration that the rule is using.
        """
        return [
            QueryConfigItem("punctuation", self.__punctuation),
        ]

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__current_state = RuleMd036States.LOOK_FOR_PARAGRAPH
        self.__start_token = None

    def __handle_look_for_parapgraph(self, token: MarkdownToken) -> RuleMd036States:
        new_state = RuleMd036States.LOOK_FOR_PARAGRAPH
        if token.is_paragraph:
            new_state = RuleMd036States.LOOK_FOR_EMPHASIS_START
            self.__start_token = token
        return new_state

    def __handle_look_for_emphasis_start(self, token: MarkdownToken) -> RuleMd036States:
        return (
            RuleMd036States.LOOK_FOR_ELIGIBLE_TEXT
            if token.is_inline_emphasis
            else RuleMd036States.LOOK_FOR_PARAGRAPH
        )

    def __handle_look_for_eligible_text(self, token: MarkdownToken) -> RuleMd036States:
        new_state = RuleMd036States.LOOK_FOR_PARAGRAPH
        if token.is_text:
            text_token = cast(TextMarkdownToken, token)
            if (
                ParserHelper.newline_character not in text_token.token_text
                and text_token.token_text[-1] not in self.__punctuation
            ):
                new_state = RuleMd036States.LOOK_FOR_EMPHASIS_END
        return new_state

    def __handle_look_for_emphasis_end(self, token: MarkdownToken) -> RuleMd036States:
        return (
            RuleMd036States.LOOK_FOR_PARAGRAPH_END
            if token.is_inline_emphasis_end
            else RuleMd036States.LOOK_FOR_PARAGRAPH
        )

    def __handle_look_for_parapgraph_end(
        self, context: PluginScanContext, token: MarkdownToken
    ) -> RuleMd036States:
        new_state = RuleMd036States.LOOK_FOR_PARAGRAPH
        if token.is_paragraph_end:
            assert self.__start_token is not None
            self.report_next_token_error(context, self.__start_token)
        return new_state

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        if self.__current_state == RuleMd036States.LOOK_FOR_PARAGRAPH:
            new_state = self.__handle_look_for_parapgraph(token)
        elif self.__current_state == RuleMd036States.LOOK_FOR_EMPHASIS_START:
            new_state = self.__handle_look_for_emphasis_start(token)
        elif self.__current_state == RuleMd036States.LOOK_FOR_ELIGIBLE_TEXT:
            new_state = self.__handle_look_for_eligible_text(token)
        elif self.__current_state == RuleMd036States.LOOK_FOR_EMPHASIS_END:
            new_state = self.__handle_look_for_emphasis_end(token)
        else:
            assert self.__current_state == RuleMd036States.LOOK_FOR_PARAGRAPH_END
            new_state = self.__handle_look_for_parapgraph_end(context, token)

        self.__current_state = new_state



"""
Module to provide a container for lines that need to be requeued.
"""

from dataclasses import dataclass
from typing import List


@dataclass(frozen=False)
class RequeueLineInfo:
    """
    Class to provide a container for lines that need to be requeued.
    """

    lines_to_requeue: List[str]
    force_ignore_first_as_lrd: bool
    force_ignore_first_as_table: bool
    has_been_abc_ed: bool = False



"""
Module to implement a plugin that looks for multiple top level headings.
"""

from typing import List, cast

from pymarkdown.extensions.front_matter_markdown_token import FrontMatterMarkdownToken
from pymarkdown.plugin_manager.plugin_details import (
    PluginDetails,
    PluginDetailsV3,
    QueryConfigItem,
)
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.atx_heading_markdown_token import AtxHeadingMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken


class RuleMd025(RulePlugin):
    """
    Class to implement a plugin that looks for multiple top level headings.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__have_top_level = False
        self.__front_matter_title: str = ""
        self.__level: int = -1

    def get_details(self) -> PluginDetails:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV3(
            plugin_name="single-title,single-h1",
            plugin_id="MD025",
            plugin_enabled_by_default=True,
            plugin_description="Multiple top-level headings in the same document",
            plugin_version="0.6.1",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md025.md",
            plugin_configuration="level, front_matter_title",
        )

    @classmethod
    def __validate_configuration_level(cls, found_value: int) -> None:
        if found_value < 1 or found_value > 6:
            raise ValueError("Allowable values are between 1 and 6.")

    @classmethod
    def __validate_configuration_title(cls, found_value: str) -> None:
        found_value = found_value.strip(" ")
        if not found_value:
            raise ValueError("Empty strings are not allowable values.")
        if ":" in found_value:
            raise ValueError("Colons (:) are not allowed in the value.")

    def initialize_from_config(self) -> None:
        """
        Event to allow the plugin to load configuration information.
        """
        self.__level = self.plugin_configuration.get_integer_property_with_default(
            "level",
            1,
            valid_value_fn=self.__validate_configuration_level,
        )
        self.__front_matter_title = (
            self.plugin_configuration.get_string_property_with_default(
                "front_matter_title",
                "title",
                valid_value_fn=self.__validate_configuration_title,
            ).lower()
        )

    def query_config(self) -> List[QueryConfigItem]:
        """
        Query to find out the configuration that the rule is using.
        """
        return [
            QueryConfigItem("level", self.__level),
            QueryConfigItem("front_matter_title", self.__front_matter_title),
        ]

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__have_top_level = False

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        if is_token_heading := token.is_atx_heading or token.is_setext_heading:
            atx_token = cast(AtxHeadingMarkdownToken, token)
            if atx_token.hash_count == self.__level and self.__have_top_level:
                self.report_next_token_error(
                    context, token, use_original_position=token.is_setext_heading
                )
            elif (atx_token.hash_count == self.__level) or (not is_token_heading):
                self.__have_top_level = True
        elif token.is_front_matter:
            front_token = cast(FrontMatterMarkdownToken, token)
            if self.__front_matter_title in front_token.matter_map:
                self.__have_top_level = True



"""
Emphasis helper
"""

import logging
from typing import List, Optional, Tuple, cast

from pymarkdown.extension_manager.extension_manager import ExtensionManager
from pymarkdown.general.constants import Constants
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.tokens.emphasis_markdown_token import EmphasisMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.special_text_markdown_token import SpecialTextMarkdownToken

POGGER = ParserLogger(logging.getLogger(__name__))


class EmphasisHelper:
    """
    Class to helper with the parsing of emphasis for inline elements.
    """

    __simple_emphasis = "*"
    __complex_emphasis = "_"
    __inline_emphasis = ""
    __strikethrough_emphasis = "~"

    @staticmethod
    def initialize(extension_manager: ExtensionManager) -> None:
        """
        Initialize this subsystem.
        """
        EmphasisHelper.__inline_emphasis = (
            f"{EmphasisHelper.__simple_emphasis}{EmphasisHelper.__complex_emphasis}"
        )
        if extension_manager.is_strike_through_enabled:
            EmphasisHelper.__inline_emphasis += EmphasisHelper.__strikethrough_emphasis

    @staticmethod
    def get_inline_emphasis() -> str:
        """
        Get the current string with all inline emphasis characters.
        """
        return EmphasisHelper.__inline_emphasis

    @staticmethod
    def __create_delimiter_stack(
        inline_blocks: List[MarkdownToken],
        is_debug_enabled: bool,
        wall_token: Optional[MarkdownToken],
    ) -> Tuple[List[SpecialTextMarkdownToken], int]:
        delimiter_stack, special_count = [], 0
        for next_block in inline_blocks:
            POGGER.debug(
                "special_count>>$>>$",
                special_count,
                next_block,
            )
            special_count += 1
            if not next_block.is_special_text:
                continue
            special_token = cast(SpecialTextMarkdownToken, next_block)
            if is_debug_enabled:
                POGGER.debug(
                    "i>>>$",
                    special_token.show_process_emphasis(),
                )
            delimiter_stack.append(special_token)

        stack_bottom = EmphasisHelper.__find_token_in_delimiter_stack(
            inline_blocks, delimiter_stack, wall_token
        )
        return delimiter_stack, stack_bottom

    @staticmethod
    def __find_potential_opener(
        delimiter_stack: List[SpecialTextMarkdownToken],
        current_position: int,
        stack_bottom: int,
        openers_bottom: int,
    ) -> Tuple[SpecialTextMarkdownToken, Optional[SpecialTextMarkdownToken]]:
        # POGGER.debug("delimiter_stack-->$",delimiter_stack)
        scan_index = current_position - 1
        found_opener: Optional[SpecialTextMarkdownToken] = None
        close_token = delimiter_stack[current_position]
        # POGGER.debug("potential closer-->$", current_position)
        while (
            scan_index >= 0
            and scan_index > stack_bottom
            and scan_index > openers_bottom
        ):
            # POGGER.debug("potential opener:$", scan_index)
            open_token = delimiter_stack[scan_index]
            if EmphasisHelper.__is_open_close_emphasis_valid(open_token, close_token):
                found_opener = open_token
                break
            scan_index -= 1
            # POGGER.debug(
            #     "scan_index-->$>stack_bottom>$>openers_bottom>$>",
            #     scan_index,
            #     stack_bottom,
            #     openers_bottom,
            # )
        return close_token, found_opener

    @staticmethod
    def __process_this_delimiter_item(
        is_debug_enabled: bool,
        delimiter_stack: List[SpecialTextMarkdownToken],
        current_position: int,
    ) -> bool:
        special_token = delimiter_stack[current_position]

        if is_debug_enabled:
            POGGER.debug(
                "Block($)-->$",
                current_position,
                special_token.show_process_emphasis(),
            )
        continue_processing = False
        if not special_token.is_active:
            POGGER.debug("not active")
        elif special_token.token_text[0] not in EmphasisHelper.get_inline_emphasis():
            POGGER.debug("not emphasis")
        elif not EmphasisHelper.__is_potential_closer(special_token):
            POGGER.debug("not closer")
        else:
            continue_processing = True
        return continue_processing

    @staticmethod
    def resolve_inline_emphasis(
        inline_blocks: List[MarkdownToken], wall_token: Optional[MarkdownToken]
    ) -> None:
        """
        Resolve the inline emphasis by interpreting the special text tokens.
        """
        is_debug_enabled = POGGER.is_debug_enabled
        delimiter_stack, stack_bottom = EmphasisHelper.__create_delimiter_stack(
            inline_blocks, is_debug_enabled, wall_token
        )

        current_position, stack_size = (
            stack_bottom + 1,
            len(delimiter_stack),
        )
        if current_position < stack_size:
            openers_bottom = stack_bottom
            if is_debug_enabled:
                POGGER.debug("BLOCK($) of ($)", current_position, stack_size)
                POGGER.debug(
                    "BLOCK($)-->$",
                    current_position,
                    delimiter_stack[current_position].show_process_emphasis(),
                )

            while current_position < (len(delimiter_stack) - 1):
                current_position += 1
                if not EmphasisHelper.__process_this_delimiter_item(
                    is_debug_enabled, delimiter_stack, current_position
                ):
                    continue

                close_token, found_opener = EmphasisHelper.__find_potential_opener(
                    delimiter_stack, current_position, stack_bottom, openers_bottom
                )

                if found_opener:
                    POGGER.debug("FOUND OPEN")
                    current_position = EmphasisHelper.__process_emphasis_pair(
                        inline_blocks,
                        found_opener,
                        close_token,
                        current_position,
                    )
                else:
                    # openers_bottom = current_position - 1
                    POGGER.debug("NOT FOUND OPEN, openers_bottom=$", openers_bottom)

                POGGER.debug("next->$", current_position)

        EmphasisHelper.__reset_token_text(inline_blocks)
        EmphasisHelper.__clear_remaining_emphasis(delimiter_stack, stack_bottom)

    # pylint: disable=too-many-arguments
    @staticmethod
    def __mark_used_tokens(
        open_token: SpecialTextMarkdownToken,
        close_token: SpecialTextMarkdownToken,
        start_index_in_blocks: int,
        end_index_in_blocks: int,
        emphasis_length: int,
        inline_blocks: List[MarkdownToken],
        current_position: int,
    ) -> int:
        # remove emphasis_length from open and close nodes
        is_debug_enabled = POGGER.is_debug_enabled
        if is_debug_enabled:
            POGGER.debug(
                "$>>close_token>>$<<",
                end_index_in_blocks,
                close_token.show_process_emphasis(),
            )
        close_token.reduce_repeat_count(emphasis_length, adjust_column_number=True)
        if not close_token.repeat_count:
            inline_blocks.remove(close_token)
            POGGER.debug("close_token>>removed")
            end_index_in_blocks -= 1
            close_token.deactivate()
        else:
            current_position -= 1
        if is_debug_enabled:
            POGGER.debug("close_token>>$<<", close_token.show_process_emphasis())
            POGGER.debug(
                "$>>open_token>>$<<",
                start_index_in_blocks,
                open_token.show_process_emphasis(),
            )
        open_token.reduce_repeat_count(emphasis_length)
        if not open_token.repeat_count:
            inline_blocks.remove(open_token)
            POGGER.debug("open_token>>removed")
            end_index_in_blocks -= 1
            open_token.deactivate()
        if is_debug_enabled:
            POGGER.debug("open_token>>$<<", open_token.show_process_emphasis())

        # "remove" between start and end from delimiter_stack
        inline_index = start_index_in_blocks + 1
        while inline_index < end_index_in_blocks:
            POGGER.debug(
                "inline_index>>$>>end>>$>>$",
                inline_index,
                end_index_in_blocks,
                len(inline_blocks),
            )
            if inline_blocks[inline_index].is_special_text:
                special_token = cast(
                    SpecialTextMarkdownToken, inline_blocks[inline_index]
                )
                special_token.deactivate()
            inline_index += 1
        return current_position

    # pylint: enable=too-many-arguments

    @staticmethod
    def __process_emphasis_pair(
        inline_blocks: List[MarkdownToken],
        open_token: SpecialTextMarkdownToken,
        close_token: SpecialTextMarkdownToken,
        current_position: int,
    ) -> int:
        """
        Given that we have found a valid open and close block, process them.
        """

        # Figure out whether we have emphasis or strong emphasis
        emphasis_character, emphasis_length = (
            open_token.token_text[0],
            2 if close_token.repeat_count >= 2 and open_token.repeat_count >= 2 else 1,
        )

        # add emph node in main stream
        POGGER.debug("open_token>>$", open_token)
        POGGER.debug("close_token>>$", close_token)

        POGGER.debug("open_token.repeat_count>>$", open_token.repeat_count)
        POGGER.debug("emphasis_length>>$", emphasis_length)
        open_column_number_delta = (
            open_token.repeat_count - emphasis_length
            if emphasis_length < open_token.repeat_count
            else 0
        )
        POGGER.debug("open_column_number_delta>>$", open_column_number_delta)

        start_index_in_blocks = inline_blocks.index(open_token)
        new_token = EmphasisMarkdownToken(
            emphasis_length,
            emphasis_character,
            line_number=open_token.line_number,
            column_number=open_token.column_number + open_column_number_delta,
        )
        inline_blocks.insert(
            start_index_in_blocks + 1,
            new_token,
        )
        end_index_in_blocks = inline_blocks.index(close_token)
        inline_blocks.insert(
            end_index_in_blocks,
            new_token.generate_close_markdown_token_from_markdown_token(
                "",
                "",
                line_number=close_token.line_number,
                column_number=close_token.column_number,
            ),
        )
        end_index_in_blocks += 1

        return EmphasisHelper.__mark_used_tokens(
            open_token,
            close_token,
            start_index_in_blocks,
            end_index_in_blocks,
            emphasis_length,
            inline_blocks,
            current_position,
        )

    @staticmethod
    def __find_token_in_delimiter_stack(
        inline_blocks: List[MarkdownToken],
        delimiter_stack: List[SpecialTextMarkdownToken],
        wall_token: Optional[MarkdownToken],
    ) -> int:
        """
        Find the specified token in the delimiter stack, based solely on
        position in the inline_blocks.
        """

        if not wall_token:
            return -1
        wall_index_in_inlines = inline_blocks.index(wall_token)
        POGGER.debug(">>wall_index_in_inlines>>$", wall_index_in_inlines)
        while wall_index_in_inlines >= 0:
            if inline_blocks[wall_index_in_inlines].is_special_text:
                special_token = cast(
                    SpecialTextMarkdownToken, inline_blocks[wall_index_in_inlines]
                )
                wall_index_in_inlines = delimiter_stack.index(special_token)
                break
            wall_index_in_inlines -= 1
        POGGER.debug(">>wall_index_in_inlines(mod)>>$", wall_index_in_inlines)
        return wall_index_in_inlines

    @staticmethod
    def __reset_token_text(inline_blocks: List[MarkdownToken]) -> None:
        """
        Once we are completed with any emphasis processing, ensure that any
        special emphasis tokens are limited to the specified lengths.
        """

        for next_block in inline_blocks:
            if next_block.is_special_text:
                special_token = cast(SpecialTextMarkdownToken, next_block)
                special_token.adjust_token_text_by_repeat_count()

    @staticmethod
    def __clear_remaining_emphasis(
        delimiter_stack: List[SpecialTextMarkdownToken], stack_bottom: int
    ) -> None:
        """
        After processing is finished, clear any active states to ensure we don't
        process them in the future.
        """

        clear_index = stack_bottom + 1
        while clear_index < len(delimiter_stack):
            delimiter_stack[clear_index].deactivate()
            clear_index += 1

    @staticmethod
    def __is_right_flanking_delimiter_run(
        current_token: SpecialTextMarkdownToken,
    ) -> bool:
        """
        Is the current token a right flanking delimiter run?
        """

        assert (
            current_token.preceding_two is not None
        ), "Preceeding character string cannot be None."
        assert (
            current_token.following_two is not None
        ), "Following character string cannot be None."
        preceding_two, following_two = (
            current_token.preceding_two.rjust(2, ParserHelper.space_character),
            current_token.following_two.ljust(2, ParserHelper.space_character),
        )

        return not Constants.unicode_whitespace.contains(preceding_two[-1]) and (
            not Constants.punctuation_characters.contains(preceding_two[-1])
            or (
                Constants.punctuation_characters.contains(preceding_two[-1])
                and (
                    Constants.unicode_whitespace.contains(following_two[0])
                    or Constants.punctuation_characters.contains(following_two[0])
                )
            )
        )

    @staticmethod
    def __is_left_flanking_delimiter_run(
        current_token: SpecialTextMarkdownToken,
    ) -> bool:
        """
        Is the current token a left flanking delimiter run?
        """
        assert (
            current_token.preceding_two is not None
        ), "Preceeding character string cannot be None."
        assert (
            current_token.following_two is not None
        ), "Following character string cannot be None."
        preceding_two, following_two = (
            current_token.preceding_two.rjust(2, ParserHelper.space_character),
            current_token.following_two.ljust(2, ParserHelper.space_character),
        )

        return not Constants.unicode_whitespace.contains(following_two[0]) and (
            not Constants.punctuation_characters.contains(following_two[0])
            or (
                Constants.punctuation_characters.contains(following_two[0])
                and (
                    Constants.unicode_whitespace.contains(preceding_two[-1])
                    or Constants.punctuation_characters.contains(preceding_two[-1])
                )
            )
        )

    @staticmethod
    def __is_potential_closer(current_token: SpecialTextMarkdownToken) -> bool:
        """
        Determine if the current token is a potential closer.
        """

        assert (
            current_token.token_text[0] in EmphasisHelper.get_inline_emphasis()
        ), "Must be a valid emphasis character."

        # Rule 3 and 7
        if current_token.token_text[0] == EmphasisHelper.__simple_emphasis:
            is_closer = EmphasisHelper.__is_right_flanking_delimiter_run(current_token)
        elif current_token.token_text[0] == EmphasisHelper.__strikethrough_emphasis:
            is_closer = (
                EmphasisHelper.__is_right_flanking_delimiter_run(current_token)
                if len(current_token.token_text) < 3
                else False
            )
        # Rule 4 and 8
        else:
            assert (
                current_token.token_text[0] == EmphasisHelper.__complex_emphasis
            ), "Must be a valid emphasis character."
            if is_closer := EmphasisHelper.__is_right_flanking_delimiter_run(
                current_token
            ):
                assert (
                    current_token.following_two is not None
                ), "Following character string cannot be None."
                is_left_flanking, following_two = (
                    EmphasisHelper.__is_left_flanking_delimiter_run(current_token),
                    current_token.following_two.ljust(2, ParserHelper.space_character),
                )
                is_closer = not is_left_flanking or (
                    is_left_flanking
                    and Constants.punctuation_characters.contains(following_two[0])
                )
        return is_closer

    @staticmethod
    def __is_potential_opener(current_token: SpecialTextMarkdownToken) -> bool:
        """
        Determine if the current token is a potential opener.
        """

        assert (
            current_token.token_text[0] in EmphasisHelper.get_inline_emphasis()
        ), "Must be a valid emphasis character."

        # Rule 1
        if current_token.token_text[0] == EmphasisHelper.__simple_emphasis:
            is_opener = EmphasisHelper.__is_left_flanking_delimiter_run(current_token)
            # POGGER.debug("is_opener (simple)=$", is_opener)
        elif current_token.token_text[0] == EmphasisHelper.__strikethrough_emphasis:
            is_opener = (
                EmphasisHelper.__is_left_flanking_delimiter_run(current_token)
                if len(current_token.token_text) < 3
                else False
            )
        else:
            assert (
                current_token.token_text[0] == EmphasisHelper.__complex_emphasis
            ), "Must be a valid emphasis character."
            is_opener = EmphasisHelper.__is_left_flanking_delimiter_run(current_token)
            # POGGER.debug("is_opener (complex)=$", is_opener)
            if is_opener:
                assert (
                    current_token.preceding_two is not None
                ), "Preceeding character string cannot be None."
                is_right_flanking, preceding_two = (
                    EmphasisHelper.__is_right_flanking_delimiter_run(current_token),
                    current_token.preceding_two.rjust(2, ParserHelper.space_character),
                )
                # POGGER.debug("is_opener (is_right_flanking)=$", is_right_flanking)
                # POGGER.debug("is_opener (preceding_two)=:$:", current_token.preceding_two)
                # POGGER.debug("is_opener (preceding_two)=:$:", preceding_two)
                is_opener = not is_right_flanking or (
                    is_right_flanking
                    and Constants.punctuation_characters.contains(preceding_two[-1])
                )
        return is_opener

    @staticmethod
    def __is_open_close_emphasis_valid(
        open_token: SpecialTextMarkdownToken, close_token: SpecialTextMarkdownToken
    ) -> bool:
        """
        Determine if these two tokens together make a valid open/close emphasis pair.
        """

        # POGGER.debug("  __is_open_close_emphasis_valid")
        # POGGER.debug("  open_token: token_text=:$:, is_active=$", open_token.token_text, open_token.is_active)
        # POGGER.debug("  close_token: token_text=:$:, is_active=$", close_token.token_text, close_token.is_active)
        is_valid_opener = False
        if not (
            open_token.token_text
            and open_token.token_text[0] == close_token.token_text[0]
        ):
            POGGER.debug("  delimiter mismatch")
        elif not open_token.is_active:
            POGGER.debug("  not active")
        else:
            is_valid_opener = EmphasisHelper.__is_potential_opener(open_token)
            POGGER.debug("is_valid_opener>>$", is_valid_opener)
            if is_valid_opener:
                is_closer_both = EmphasisHelper.__is_potential_closer(
                    close_token
                ) and EmphasisHelper.__is_potential_opener(close_token)
                # POGGER.debug("is_closer_both>>$", is_closer_both)
                is_opener_both = EmphasisHelper.__is_potential_closer(
                    open_token
                ) and EmphasisHelper.__is_potential_opener(open_token)
                # POGGER.debug("is_opener_both>>$", is_opener_both)
                if is_closer_both or is_opener_both:
                    sum_repeat_count = (
                        close_token.repeat_count + open_token.repeat_count
                    )
                    # POGGER.debug("sum_delims>>$", sum_repeat_count)
                    # POGGER.debug("closer_delims>>$", close_token.repeat_count)
                    # POGGER.debug("opener_delims>>$", open_token.repeat_count)

                    if sum_repeat_count % 3 == 0:
                        is_valid_opener = (
                            close_token.repeat_count % 3 == 0
                            and open_token.repeat_count % 3 == 0
                        )

        return is_valid_opener



"""
Module to hold the tuple of information for creating a Link Reference Definition.
"""

from dataclasses import dataclass
from typing import Optional

from pymarkdown.links.link_reference_info import LinkReferenceInfo
from pymarkdown.links.link_reference_titles import LinkReferenceTitles


@dataclass(frozen=True)
class LinkReferenceDefinitionTuple:
    """
    Class to hold the tuple of information for creating a Link Reference Definition.
    """

    normalized_destination: Optional[str]
    link_titles: LinkReferenceTitles
    link_info: LinkReferenceInfo
